"""
unit tests for the grains state
"""

import contextlib
import os

import pytest

import salt.modules.grains as grainsmod
import salt.states.grains as grains
import salt.utils.files
import salt.utils.stringutils
import salt.utils.yaml
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    minion_opts["local"] = True
    minion_opts["test"] = False
    loader_globals = {
        "__opts__": minion_opts,
        "__salt__": {
            "cmd.run_all": MagicMock(
                return_value={"pid": 5, "retcode": 0, "stderr": "", "stdout": ""}
            ),
            "grains.get": grainsmod.get,
            "grains.set": grainsmod.set,
            "grains.setval": grainsmod.setval,
            "grains.delval": grainsmod.delval,
            "grains.append": grainsmod.append,
            "grains.remove": grainsmod.remove,
            "saltutil.sync_grains": MagicMock(),
        },
    }
    return {grains: loader_globals, grainsmod: loader_globals}


def assert_grain_file_content(grains_string):
    grains_file = os.path.join(grains.__opts__["conf_dir"], "grains")
    with salt.utils.files.fopen(grains_file, "r") as grf:
        grains_data = salt.utils.stringutils.to_unicode(grf.read())
    assert grains_string == grains_data


@contextlib.contextmanager
def set_grains(grains_data):
    with patch.dict(grains.__grains__, grains_data):
        with patch.dict(grainsmod.__grains__, grains_data):
            grains_file = os.path.join(grains.__opts__["conf_dir"], "grains")
            with salt.utils.files.fopen(grains_file, "w+") as grf:
                salt.utils.yaml.safe_dump(grains_data, grf, default_flow_style=False)
            yield


# 'exists' function tests: 2


def test_exists_missing():
    with set_grains({"a": "aval"}):
        ret = grains.exists(name="foo")
        assert ret["result"] is False
        assert ret["comment"] == "Grain does not exist"
        assert ret["changes"] == {}


def test_exists_found():
    with set_grains({"a": "aval", "foo": "bar"}):
        # Grain already set
        ret = grains.exists(name="foo")
        assert ret["result"] is True
        assert ret["comment"] == "Grain exists"
        assert ret["changes"] == {}

    # 'make_hashable' function tests: 1


def test_make_hashable():
    with set_grains({"cmplx_lst_grain": [{"a": "aval"}, {"foo": "bar"}]}):
        hashable_list = {"cmplx_lst_grain": [{"a": "aval"}, {"foo": "bar"}]}
        assert grains.make_hashable(grains.__grains__).issubset(
            grains.make_hashable(hashable_list)
        )

    # 'present' function tests: 12


def test_present_add():
    # Set a non existing grain
    with set_grains({"a": "aval"}):
        ret = grains.present(name="foo", value="bar")
        assert ret["result"] is True
        assert ret["changes"] == {"foo": "bar"}
        assert grains.__grains__ == {"a": "aval", "foo": "bar"}
        assert_grain_file_content("a: aval\nfoo: bar\n")

    # Set a non existing nested grain
    with set_grains({"a": "aval"}):
        ret = grains.present(name="foo:is:nested", value="bar")
        assert ret["result"] is True
        assert ret["changes"] == {"foo": {"is": {"nested": "bar"}}}
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": "bar"}}}
        assert_grain_file_content("a: aval\nfoo:\n  is:\n    nested: bar\n")

    # Set a non existing nested dict grain
    with set_grains({"a": "aval"}):
        ret = grains.present(name="foo:is:nested", value={"bar": "is a dict"})
        assert ret["result"] is True
        assert ret["changes"] == {"foo": {"is": {"nested": {"bar": "is a dict"}}}}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": {"is": {"nested": {"bar": "is a dict"}}},
        }
        assert_grain_file_content(
            "a: aval\nfoo:\n  is:\n    nested:\n      bar: is a dict\n"
        )


def test_present_add_key_to_existing():
    with set_grains({"a": "aval", "foo": {"k1": "v1"}}):
        # Fails setting a grain to a dict
        ret = grains.present(name="foo:k2", value="v2")
        assert ret["result"] is True
        assert ret["comment"] == "Set grain foo:k2 to v2"
        assert ret["changes"] == {"foo": {"k2": "v2", "k1": "v1"}}
        assert grains.__grains__ == {"a": "aval", "foo": {"k1": "v1", "k2": "v2"}}
        assert_grain_file_content("a: aval\nfoo:\n  k1: v1\n  k2: v2\n")


def test_present_already_set():
    with set_grains({"a": "aval", "foo": "bar"}):
        # Grain already set
        ret = grains.present(name="foo", value="bar")
        assert ret["result"] is True
        assert ret["comment"] == "Grain is already set"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": "bar"}

    with set_grains({"a": "aval", "foo": {"is": {"nested": "bar"}}}):
        # Nested grain already set
        ret = grains.present(name="foo:is:nested", value="bar")
        assert ret["result"] is True
        assert ret["comment"] == "Grain is already set"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": "bar"}}}

    with set_grains({"a": "aval", "foo": {"is": {"nested": "bar"}}}):
        # Nested dict grain already set
        ret = grains.present(name="foo:is", value={"nested": "bar"})
        assert ret["result"] is True
        assert ret["comment"] == "Grain is already set"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": "bar"}}}


def test_present_overwrite():
    with set_grains({"a": "aval", "foo": "bar"}):
        # Overwrite an existing grain
        ret = grains.present(name="foo", value="newbar")
        assert ret["result"] is True
        assert ret["changes"] == {"foo": "newbar"}
        assert grains.__grains__ == {"a": "aval", "foo": "newbar"}
        assert_grain_file_content("a: aval\nfoo: newbar\n")

    with set_grains({"a": "aval", "foo": "bar"}):
        # Clear a grain (set to None)
        ret = grains.present(name="foo", value=None)
        assert ret["result"] is True
        assert ret["changes"] == {"foo": None}
        assert grains.__grains__ == {"a": "aval", "foo": None}
        assert_grain_file_content("a: aval\nfoo: null\n")

    with set_grains({"a": "aval", "foo": {"is": {"nested": "bar"}}}):
        # Overwrite an existing nested grain
        ret = grains.present(name="foo:is:nested", value="newbar")
        assert ret["result"] is True
        assert ret["changes"] == {"foo": {"is": {"nested": "newbar"}}}
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": "newbar"}}}
        assert_grain_file_content("a: aval\nfoo:\n  is:\n    nested: newbar\n")

    with set_grains({"a": "aval", "foo": {"is": {"nested": "bar"}}}):
        # Clear a nested grain (set to None)
        ret = grains.present(name="foo:is:nested", value=None)
        assert ret["result"] is True
        assert ret["changes"] == {"foo": {"is": {"nested": None}}}
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": None}}}
        assert_grain_file_content("a: aval\nfoo:\n  is:\n    nested: null\n")


def test_present_fail_overwrite():
    with set_grains({"a": "aval", "foo": {"is": {"nested": "val"}}}):
        # Overwrite an existing grain
        ret = grains.present(name="foo:is", value="newbar")
        assert ret["result"] is False
        assert ret["changes"] == {}
        assert (
            ret["comment"]
            == "The key 'foo:is' exists but is a dict or a list. Use 'force=True' to overwrite."
        )
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": "val"}}}

    with set_grains({"a": "aval", "foo": {"is": {"nested": "val"}}}):
        # Clear a grain (set to None)
        ret = grains.present(name="foo:is", value=None)
        assert ret["result"] is False
        assert ret["changes"] == {}
        assert (
            ret["comment"]
            == "The key 'foo:is' exists but is a dict or a list. Use 'force=True' to overwrite."
        )
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": "val"}}}


def test_present_fails_to_set_dict_or_list():
    with set_grains({"a": "aval", "foo": "bar"}):
        # Fails to overwrite a grain to a list
        ret = grains.present(name="foo", value=["l1", "l2"])
        assert ret["result"] is False
        assert (
            ret["comment"]
            == "The key 'foo' exists and the given value is a dict or a list. Use 'force=True' to overwrite."
        )
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": "bar"}

    with set_grains({"a": "aval", "foo": "bar"}):
        # Fails setting a grain to a dict
        ret = grains.present(name="foo", value={"k1": "v1"})
        assert ret["result"] is False
        assert (
            ret["comment"]
            == "The key 'foo' exists and the given value is a dict or a list. Use 'force=True' to overwrite."
        )
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": "bar"}

    with set_grains({"a": "aval", "foo": {"is": {"nested": "bar"}}}):
        # Fails to overwrite a nested grain to a list
        ret = grains.present(name="foo,is,nested", value=["l1", "l2"], delimiter=",")
        assert ret["result"] is False
        assert ret["changes"] == {}
        assert (
            ret["comment"]
            == "The key 'foo:is:nested' exists and the given value is a dict or a list. Use 'force=True' to overwrite."
        )
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": "bar"}}}

    with set_grains({"a": "aval", "foo": {"is": {"nested": "bar"}}}):
        # Fails setting a nested grain to a dict
        ret = grains.present(name="foo:is:nested", value={"k1": "v1"})
        assert ret["result"] is False
        assert (
            ret["comment"]
            == "The key 'foo:is:nested' exists and the given value is a dict or a list. Use 'force=True' to overwrite."
        )
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": "bar"}}}


def test_present_fail_merge_dict():
    with set_grains({"a": "aval", "foo": {"k1": "v1"}}):
        # Fails setting a grain to a dict
        ret = grains.present(name="foo", value={"k2": "v2"})
        assert ret["result"] is False
        assert (
            ret["comment"]
            == "The key 'foo' exists but is a dict or a list. Use 'force=True' to overwrite."
        )
        assert grains.__grains__ == {"a": "aval", "foo": {"k1": "v1"}}
        assert_grain_file_content("a: aval\nfoo:\n  k1: v1\n")


def test_present_force_to_set_dict_or_list():
    with set_grains({"a": "aval", "foo": "bar"}):
        # Force to overwrite a grain to a list
        ret = grains.present(name="foo", value=["l1", "l2"], force=True)
        assert ret["result"] is True
        assert ret["comment"] == "Set grain foo to ['l1', 'l2']"
        assert ret["changes"] == {"foo": ["l1", "l2"]}
        assert grains.__grains__ == {"a": "aval", "foo": ["l1", "l2"]}
        assert_grain_file_content("a: aval\nfoo:\n- l1\n- l2\n")

    with set_grains({"a": "aval", "foo": "bar"}):
        # Force setting a grain to a dict
        ret = grains.present(name="foo", value={"k1": "v1"}, force=True)
        assert ret["result"] is True
        assert ret["comment"] == "Set grain foo to {'k1': 'v1'}"
        assert ret["changes"] == {"foo": {"k1": "v1"}}
        assert grains.__grains__ == {"a": "aval", "foo": {"k1": "v1"}}
        assert_grain_file_content("a: aval\nfoo:\n  k1: v1\n")

    with set_grains({"a": "aval", "foo": {"is": {"nested": "bar"}}}):
        # Force to overwrite a nested grain to a list
        ret = grains.present(
            name="foo,is,nested", value=["l1", "l2"], delimiter=",", force=True
        )
        assert ret["result"] is True
        assert ret["changes"] == {"foo": {"is": {"nested": ["l1", "l2"]}}}
        assert ret["comment"] == "Set grain foo:is:nested to ['l1', 'l2']"
        assert grains.__grains__ == {
            "a": "aval",
            "foo": {"is": {"nested": ["l1", "l2"]}},
        }
        assert_grain_file_content(
            "a: aval\nfoo:\n  is:\n    nested:\n    - l1\n    - l2\n"
        )

    with set_grains({"a": "aval", "foo": {"is": {"nested": "bar"}, "and": "other"}}):
        # Force setting a nested grain to a dict
        ret = grains.present(name="foo:is:nested", value={"k1": "v1"}, force=True)
        assert ret["result"] is True
        assert ret["comment"] == "Set grain foo:is:nested to {'k1': 'v1'}"
        assert ret["changes"] == {
            "foo": {"is": {"nested": {"k1": "v1"}}, "and": "other"}
        }
        assert grains.__grains__ == {
            "a": "aval",
            "foo": {"is": {"nested": {"k1": "v1"}}, "and": "other"},
        }
        assert_grain_file_content(
            "a: aval\nfoo:\n  and: other\n  is:\n    nested:\n      k1: v1\n"
        )


def test_present_fails_to_convert_value_to_key():
    with set_grains({"a": "aval", "foo": "bar"}):
        # Fails converting a value to a nested grain key
        ret = grains.present(name="foo:is:nested", value={"k1": "v1"})
        assert ret["result"] is False
        assert (
            ret["comment"]
            == "The key 'foo' value is 'bar', which is different from the provided key 'is'. Use 'force=True' to overwrite."
        )
        assert ret["changes"] == {}


def test_present_overwrite_test():
    with patch.dict(grains.__opts__, {"test": True}):
        with set_grains({"a": "aval", "foo": "bar"}):
            # Overwrite an existing grain
            ret = grains.present(name="foo", value="newbar")
            assert ret["result"] is None
            assert ret["changes"] == {"changed": {"foo": "newbar"}}
            assert grains.__grains__ == {"a": "aval", "foo": "bar"}
            assert_grain_file_content("a: aval\nfoo: bar\n")


def test_present_convert_value_to_key():
    with set_grains({"a": "aval", "foo": "is"}):
        # Converts a value to a nested grain key
        ret = grains.present(name="foo:is:nested", value={"k1": "v1"})
        assert ret["result"] is True
        assert ret["comment"] == "Set grain foo:is:nested to {'k1': 'v1'}"
        assert ret["changes"] == {"foo": {"is": {"nested": {"k1": "v1"}}}}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": {"is": {"nested": {"k1": "v1"}}},
        }
        assert_grain_file_content("a: aval\nfoo:\n  is:\n    nested:\n      k1: v1\n")

    with set_grains({"a": "aval", "foo": ["one", "is", "correct"]}):
        # Converts a list element to a nested grain key
        ret = grains.present(name="foo:is:nested", value={"k1": "v1"})
        assert ret["result"] is True
        assert ret["comment"] == "Set grain foo:is:nested to {'k1': 'v1'}"
        assert ret["changes"] == {
            "foo": ["one", {"is": {"nested": {"k1": "v1"}}}, "correct"]
        }
        assert grains.__grains__ == {
            "a": "aval",
            "foo": ["one", {"is": {"nested": {"k1": "v1"}}}, "correct"],
        }
        assert_grain_file_content(
            "a: aval\nfoo:\n- one\n- is:\n    nested:\n      k1: v1\n- correct\n"
        )


def test_present_unknown_failure():
    with patch("salt.modules.grains.setval") as mocked_setval:
        mocked_setval.return_value = "Failed to set grain foo"
        with set_grains({"a": "aval", "foo": "bar"}):
            # Unknown reason failure
            ret = grains.present(name="foo", value="baz")
            assert ret["result"] is False
            assert ret["comment"] == "Failed to set grain foo"
            assert ret["changes"] == {}
            assert grains.__grains__ == {"a": "aval", "foo": "bar"}
            assert_grain_file_content("a: aval\nfoo: bar\n")


# 'absent' function tests: 6


def test_absent_already():
    # Unset a non existent grain
    with set_grains({"a": "aval"}):
        ret = grains.absent(name="foo")
        assert ret["result"] is True
        assert ret["comment"] == "Grain foo does not exist"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval"}
        assert_grain_file_content("a: aval\n")

    # Unset a non existent nested grain
    with set_grains({"a": "aval"}):
        ret = grains.absent(name="foo:is:nested")
        assert ret["result"] is True
        assert ret["comment"] == "Grain foo:is:nested does not exist"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval"}
        assert_grain_file_content("a: aval\n")


def test_absent_unset():
    # Unset a grain
    with set_grains({"a": "aval", "foo": "bar"}):
        ret = grains.absent(name="foo")
        assert ret["result"] is True
        assert ret["comment"] == "Value for grain foo was set to None"
        assert ret["changes"] == {"grain": "foo", "value": None}
        assert grains.__grains__ == {"a": "aval", "foo": None}
        assert_grain_file_content("a: aval\nfoo: null\n")

    # Unset grain when its value is False
    with set_grains({"a": "aval", "foo": False}):
        ret = grains.absent(name="foo")
        assert ret["result"] is True
        assert ret["comment"] == "Value for grain foo was set to None"
        assert ret["changes"] == {"grain": "foo", "value": None}
        assert grains.__grains__ == {"a": "aval", "foo": None}
        assert_grain_file_content("a: aval\nfoo: null\n")

    # Unset a nested grain
    with set_grains(
        {"a": "aval", "foo": ["order", {"is": {"nested": "bar"}}, "correct"]}
    ):
        ret = grains.absent(name="foo,is,nested", delimiter=",")
        assert ret["result"] is True
        assert ret["comment"] == "Value for grain foo:is:nested was set to None"
        assert ret["changes"] == {"grain": "foo:is:nested", "value": None}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": ["order", {"is": {"nested": None}}, "correct"],
        }
        assert_grain_file_content(
            "a: aval\nfoo:\n- order\n- is:\n    nested: null\n- correct\n"
        )

    # Unset a nested value don't change anything
    with set_grains({"a": "aval", "foo": ["order", {"is": "nested"}, "correct"]}):
        ret = grains.absent(name="foo:is:nested")
        assert ret["result"] is True
        assert ret["comment"] == "Grain foo:is:nested does not exist"
        assert ret["changes"] == {}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": ["order", {"is": "nested"}, "correct"],
        }
        assert_grain_file_content("a: aval\nfoo:\n- order\n- is: nested\n- correct\n")


def test_absent_unset_test():
    with patch.dict(grains.__opts__, {"test": True}):
        with set_grains({"a": "aval", "foo": "bar"}):
            # Overwrite an existing grain
            ret = grains.absent(name="foo")
            assert ret["result"] is None
            assert ret["changes"] == {"grain": "foo", "value": None}
            assert grains.__grains__ == {"a": "aval", "foo": "bar"}
            assert_grain_file_content("a: aval\nfoo: bar\n")


def test_absent_fails_nested_complex_grain():
    # Unset a nested complex grain
    with set_grains(
        {"a": "aval", "foo": ["order", {"is": {"nested": "bar"}}, "correct"]}
    ):
        ret = grains.absent(name="foo:is")
        assert ret["result"] is False
        assert (
            ret["comment"]
            == "The key 'foo:is' exists but is a dict or a list. Use 'force=True' to overwrite."
        )
        assert ret["changes"] == {}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": ["order", {"is": {"nested": "bar"}}, "correct"],
        }
        assert_grain_file_content(
            "a: aval\nfoo:\n- order\n- is:\n    nested: bar\n- correct\n"
        )


def test_absent_force_nested_complex_grain():
    # Unset a nested complex grain
    with set_grains(
        {"a": "aval", "foo": ["order", {"is": {"nested": "bar"}}, "correct"]}
    ):
        ret = grains.absent(name="foo:is", force=True)
        assert ret["result"] is True
        assert ret["comment"] == "Value for grain foo:is was set to None"
        assert ret["changes"] == {"grain": "foo:is", "value": None}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": ["order", {"is": None}, "correct"],
        }
        assert_grain_file_content("a: aval\nfoo:\n- order\n- is: null\n- correct\n")


def test_absent_delete():
    # Delete a grain
    with set_grains({"a": "aval", "foo": "bar"}):
        ret = grains.absent(name="foo", destructive=True)
        assert ret["result"] is True
        assert ret["comment"] == "Grain foo was deleted"
        assert ret["changes"] == {"deleted": "foo"}
        assert grains.__grains__ == {"a": "aval"}
        assert_grain_file_content("a: aval\n")

    # Delete a previously unset grain
    with set_grains({"a": "aval", "foo": None}):
        ret = grains.absent(name="foo", destructive=True)
        assert ret["result"] is True
        assert ret["comment"] == "Grain foo was deleted"
        assert ret["changes"] == {"deleted": "foo"}
        assert grains.__grains__ == {"a": "aval"}
        assert_grain_file_content("a: aval\n")

    # Delete a nested grain
    with set_grains(
        {
            "a": "aval",
            "foo": [
                "order",
                {"is": {"nested": "bar", "other": "value"}},
                "correct",
            ],
        }
    ):
        ret = grains.absent(name="foo:is:nested", destructive=True)
        assert ret["result"] is True
        assert ret["comment"] == "Grain foo:is:nested was deleted"
        assert ret["changes"] == {"deleted": "foo:is:nested"}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": ["order", {"is": {"other": "value"}}, "correct"],
        }
        assert_grain_file_content(
            "a: aval\nfoo:\n- order\n- is:\n    other: value\n- correct\n"
        )


# 'append' function tests: 6


def test_append():
    # Append to an existing list
    with set_grains({"a": "aval", "foo": ["bar"]}):
        ret = grains.append(name="foo", value="baz")
        assert ret["result"] is True
        assert ret["comment"] == "Value baz was added to grain foo"
        assert ret["changes"] == {"added": "baz"}
        assert grains.__grains__ == {"a": "aval", "foo": ["bar", "baz"]}
        assert_grain_file_content("a: aval\nfoo:\n- bar\n- baz\n")


def test_append_nested():
    # Append to an existing nested list
    with set_grains({"a": "aval", "foo": {"list": ["bar"]}}):
        ret = grains.append(name="foo,list", value="baz", delimiter=",")
        assert ret["result"] is True
        assert ret["comment"] == "Value baz was added to grain foo:list"
        assert ret["changes"] == {"added": "baz"}
        assert grains.__grains__ == {"a": "aval", "foo": {"list": ["bar", "baz"]}}
        assert_grain_file_content("a: aval\nfoo:\n  list:\n  - bar\n  - baz\n")


def test_append_already():
    # Append to an existing list
    with set_grains({"a": "aval", "foo": ["bar"]}):
        ret = grains.append(name="foo", value="bar")
        assert ret["result"] is True
        assert ret["comment"] == "Value bar is already in the list " + "for grain foo"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": ["bar"]}
        assert_grain_file_content("a: aval\nfoo:\n- bar\n")


def test_append_fails_not_a_list():
    # Fail to append to an existing grain, not a list
    with set_grains({"a": "aval", "foo": {"bar": "val"}}):
        ret = grains.append(name="foo", value="baz")
        assert ret["result"] is False
        assert ret["comment"] == "Grain foo is not a valid list"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": {"bar": "val"}}


def test_append_convert_to_list():
    # Append to an existing grain, converting to a list
    with set_grains({"a": "aval", "foo": {"bar": "val"}}):
        assert_grain_file_content("a: aval\nfoo:\n  bar: val\n")
        ret = grains.append(name="foo", value="baz", convert=True)
        assert ret["result"] is True
        assert ret["comment"] == "Value baz was added to grain foo"
        assert ret["changes"] == {"added": "baz"}
        assert grains.__grains__ == {"a": "aval", "foo": [{"bar": "val"}, "baz"]}
        assert_grain_file_content("a: aval\nfoo:\n- bar: val\n- baz\n")

    # Append to an existing grain, converting to a list a multi-value dict
    with set_grains({"a": "aval", "foo": {"bar": "val", "other": "value"}}):
        assert_grain_file_content("a: aval\nfoo:\n  bar: val\n  other: value\n")
        ret = grains.append(name="foo", value="baz", convert=True)
        assert ret["result"] is True
        assert ret["comment"] == "Value baz was added to grain foo"
        assert ret["changes"] == {"added": "baz"}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": [{"bar": "val", "other": "value"}, "baz"],
        }
        assert_grain_file_content("a: aval\nfoo:\n- bar: val\n  other: value\n- baz\n")


def test_append_fails_inexistent():
    # Append to a non existing grain
    with set_grains({"a": "aval"}):
        ret = grains.append(name="foo", value="bar")
        assert ret["result"] is False
        assert ret["comment"] == "Grain foo does not exist"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval"}


def test_append_convert_to_list_empty():
    # Append to an existing list
    with set_grains({"foo": None}):
        ret = grains.append(name="foo", value="baz", convert=True)
        assert ret["result"] is True
        assert ret["comment"] == "Value baz was added to grain foo"
        assert ret["changes"] == {"added": "baz"}
        assert grains.__grains__ == {"foo": ["baz"]}
        assert_grain_file_content("foo:\n- baz\n")


# 'list_present' function tests: 7


def test_list_present():
    with set_grains({"a": "aval", "foo": ["bar"]}):
        ret = grains.list_present(name="foo", value="baz")
        assert ret["result"] is True
        assert ret["comment"] == "Append value baz to grain foo"
        assert ret["changes"] == {"new": {"foo": ["bar", "baz"]}}
        assert grains.__grains__ == {"a": "aval", "foo": ["bar", "baz"]}
        assert_grain_file_content("a: aval\nfoo:\n- bar\n- baz\n")


def test_list_present_nested():
    with set_grains({"a": "aval", "foo": {"is": {"nested": ["bar"]}}}):
        ret = grains.list_present(name="foo,is,nested", value="baz", delimiter=",")
        assert ret["result"] is True
        assert ret["comment"] == "Append value baz to grain foo:is:nested"
        assert ret["changes"] == {"new": {"foo": {"is": {"nested": ["bar", "baz"]}}}}
        assert grains.__grains__ == {
            "a": "aval",
            "foo": {"is": {"nested": ["bar", "baz"]}},
        }
        assert_grain_file_content(
            "a: aval\nfoo:\n  is:\n    nested:\n    - bar\n    - baz\n"
        )


def test_list_present_inexistent():
    with set_grains({"a": "aval"}):
        ret = grains.list_present(name="foo", value="baz")
        assert ret["result"] is True
        assert ret["comment"] == "Append value baz to grain foo"
        assert ret["changes"] == {"new": {"foo": ["baz"]}}
        assert grains.__grains__ == {"a": "aval", "foo": ["baz"]}
        assert_grain_file_content("a: aval\nfoo:\n- baz\n")


def test_list_present_inexistent_nested():
    with set_grains({"a": "aval"}):
        ret = grains.list_present(name="foo:is:nested", value="baz")
        assert ret["result"] is True
        assert ret["comment"] == "Append value baz to grain foo:is:nested"
        assert ret["changes"] == {"new": {"foo": {"is": {"nested": ["baz"]}}}}
        assert grains.__grains__ == {"a": "aval", "foo": {"is": {"nested": ["baz"]}}}
        assert_grain_file_content("a: aval\nfoo:\n  is:\n    nested:\n    - baz\n")


def test_list_present_not_a_list():
    with set_grains({"a": "aval", "foo": "bar"}):
        ret = grains.list_present(name="foo", value="baz")
        assert ret["result"] is False
        assert ret["comment"] == "Grain foo is not a valid list"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": "bar"}
        assert_grain_file_content("a: aval\nfoo: bar\n")


def test_list_present_nested_already():
    with set_grains({"a": "aval", "b": {"foo": ["bar"]}}):
        ret = grains.list_present(name="b:foo", value="bar")
        assert ret["result"] is True
        assert ret["comment"] == "Value bar is already in grain b:foo"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "b": {"foo": ["bar"]}}
        assert_grain_file_content("a: aval\nb:\n  foo:\n  - bar\n")


def test_list_present_already():
    with set_grains({"a": "aval", "foo": ["bar"]}):
        ret = grains.list_present(name="foo", value="bar")
        assert ret["result"] is True
        assert ret["comment"] == "Value bar is already in grain foo"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": ["bar"]}
        assert_grain_file_content("a: aval\nfoo:\n- bar\n")


def test_list_present_unknown_failure():
    with set_grains({"a": "aval", "foo": ["bar"]}):
        # Unknown reason failure

        with patch.dict(grainsmod.__salt__, {"grains.append": MagicMock()}):
            ret = grains.list_present(name="foo", value="baz")
            assert ret["result"] is False
            assert ret["comment"] == "Failed append value baz to grain foo"
            assert ret["changes"] == {}
            assert grains.__grains__ == {"a": "aval", "foo": ["bar"]}
            assert_grain_file_content("a: aval\nfoo:\n- bar\n")


# 'list_absent' function tests: 6


def test_list_absent():
    with set_grains({"a": "aval", "foo": ["bar"]}):
        ret = grains.list_absent(name="foo", value="bar")
        assert ret["result"] is True
        assert ret["comment"] == "Value bar was deleted from grain foo"
        assert ret["changes"] == {"deleted": ["bar"]}
        assert grains.__grains__ == {"a": "aval", "foo": []}
        assert_grain_file_content("a: aval\nfoo: []\n")


def test_list_absent_nested():
    with set_grains({"a": "aval", "foo": {"list": ["bar"]}}):
        ret = grains.list_absent(name="foo:list", value="bar")
        assert ret["result"] is True
        assert ret["comment"] == "Value bar was deleted from grain foo:list"
        assert ret["changes"] == {"deleted": ["bar"]}
        assert grains.__grains__ == {"a": "aval", "foo": {"list": []}}
        assert_grain_file_content("a: aval\nfoo:\n  list: []\n")


def test_list_absent_inexistent():
    with set_grains({"a": "aval"}):
        ret = grains.list_absent(name="foo", value="baz")
        assert ret["result"] is True
        assert ret["comment"] == "Grain foo does not exist"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval"}
        assert_grain_file_content("a: aval\n")


def test_list_absent_inexistent_nested():
    with set_grains({"a": "aval"}):
        ret = grains.list_absent(name="foo:list", value="baz")
        assert ret["result"] is True
        assert ret["comment"] == "Grain foo:list does not exist"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval"}
        assert_grain_file_content("a: aval\n")


def test_list_absent_not_a_list():
    with set_grains({"a": "aval", "foo": "bar"}):
        ret = grains.list_absent(name="foo", value="bar")
        assert ret["result"] is False
        assert ret["comment"] == "Grain foo is not a valid list"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": "bar"}
        assert_grain_file_content("a: aval\nfoo: bar\n")


def test_list_absent_already():
    with set_grains({"a": "aval", "foo": ["bar"]}):
        ret = grains.list_absent(name="foo", value="baz")
        assert ret["result"] is True
        assert ret["comment"] == "Value baz is absent from grain foo"
        assert ret["changes"] == {}
        assert grains.__grains__ == {"a": "aval", "foo": ["bar"]}
        assert_grain_file_content("a: aval\nfoo:\n- bar\n")



"""
A route is a rule that specifies how certain packets should be handled by the
virtual network. Routes are associated with virtual machine instances by tag,
and the set of routes for a particular VM is called its routing table.
For each packet leaving a virtual machine, the system searches that machine's
routing table for a single best matching route.

.. versionadded:: 2018.3.0

This module will create a route to send traffic destined to the Internet
through your gateway instance.

:codeauthor: `Pratik Bandarkar <pratik.bandarkar@gmail.com>`
:maturity:   new
:depends:    google-api-python-client
:platform:   Linux

"""

import logging

try:
    import googleapiclient.discovery
    import oauth2client.service_account

    HAS_LIB = True
except ImportError:
    HAS_LIB = False

log = logging.getLogger(__name__)

__virtualname__ = "gcp"


def __virtual__():
    """
    Check for googleapiclient api
    """
    if HAS_LIB is False:
        return (
            False,
            "Required dependencies 'googleapiclient' and/or 'oauth2client' were not"
            " found.",
        )
    return __virtualname__


def _get_network(project_id, network_name, service):
    """
    Fetch network selfLink from network name.
    """
    return service.networks().get(project=project_id, network=network_name).execute()


def _get_instance(project_id, instance_zone, name, service):
    """
    Get instance details
    """
    return (
        service.instances()
        .get(project=project_id, zone=instance_zone, instance=name)
        .execute()
    )


def route_create(
    credential_file=None,
    project_id=None,
    name=None,
    dest_range=None,
    next_hop_instance=None,
    instance_zone=None,
    tags=None,
    network=None,
    priority=None,
):
    """
    Create a route to send traffic destined to the Internet through your
    gateway instance

    credential_file : string
        File location of application default credential. For more information,
        refer: https://developers.google.com/identity/protocols/application-default-credentials
    project_id : string
        Project ID where instance and network resides.
    name : string
        name of the route to create
    next_hop_instance : string
        the name of an instance that should handle traffic matching this route.
    instance_zone : string
        zone where instance("next_hop_instance") resides
    network : string
        Specifies the network to which the route will be applied.
    dest_range : string
        The destination range of outgoing packets that the route will apply to.
    tags : list
        (optional) Identifies the set of instances that this route will apply to.
    priority : int
        (optional) Specifies the priority of this route relative to other routes.
        default=1000

    CLI Example:

    .. code-block:: bash

        salt 'salt-master.novalocal' gcp.route_create
            credential_file=/root/secret_key.json
            project_id=cp100-170315
            name=derby-db-route1
            next_hop_instance=instance-1
            instance_zone=us-central1-a
            network=default
            dest_range=0.0.0.0/0
            tags=['no-ip']
            priority=700

    In above example, the instances which are having tag "no-ip" will route the
    packet to instance "instance-1"(if packet is intended to other network)
    """

    credentials = (
        oauth2client.service_account.ServiceAccountCredentials.from_json_keyfile_name(
            credential_file
        )
    )
    service = googleapiclient.discovery.build("compute", "v1", credentials=credentials)
    routes = service.routes()

    routes_config = {
        "name": str(name),
        "network": _get_network(project_id, str(network), service=service)["selfLink"],
        "destRange": str(dest_range),
        "nextHopInstance": _get_instance(
            project_id, instance_zone, next_hop_instance, service=service
        )["selfLink"],
        "tags": tags,
        "priority": priority,
    }
    route_create_request = routes.insert(project=project_id, body=routes_config)
    return route_create_request.execute()



"""
A module that wraps calls to the Chocolatey package manager
(http://chocolatey.org)

.. versionadded:: 2014.1.0
"""

import logging
import os
import re
import tempfile

from requests.structures import CaseInsensitiveDict

import salt.utils.data
import salt.utils.platform
from salt.exceptions import (
    CommandExecutionError,
    CommandNotFoundError,
    MinionError,
    SaltInvocationError,
)
from salt.utils.data import CaseInsensitiveDict
from salt.utils.versions import Version

log = logging.getLogger(__name__)

__func_alias__ = {"list_": "list"}

__virtualname__ = "chocolatey"


def __virtual__():
    """
    Confirm this module is on a Windows system running Vista or later.

    While it is possible to make Chocolatey run under XP and Server 2003 with
    an awful lot of hassle (e.g. SSL is completely broken), the PowerShell shim
    for simulating UAC forces a GUI prompt, and is not compatible with
    salt-minion running as SYSTEM.
    """
    if not salt.utils.platform.is_windows():
        return False, "Chocolatey: Requires Windows"

    if __grains__["osrelease"] in ("XP", "2003Server"):
        return False, "Chocolatey: Requires Windows Vista or later"

    return __virtualname__


def _clear_context():
    """
    Clear variables stored in __context__. Run this function when a new version
    of chocolatey is installed.
    """
    choco_items = [x for x in __context__ if x.startswith("chocolatey.")]
    for var in choco_items:
        __context__.pop(var)


def _yes():
    """
    Returns ['--yes'] if on v0.9.9.0 or later, otherwise returns an empty list
    Confirm all prompts (--yes_ is available on v0.9.9.0 or later
    """
    if "chocolatey._yes" in __context__:
        return __context__["chocolatey._yes"]
    if Version(chocolatey_version()) >= Version("0.9.9"):
        answer = ["--yes"]
    else:
        answer = []
    __context__["chocolatey._yes"] = answer
    return __context__["chocolatey._yes"]


def _no_progress():
    """
    Returns ['--no-progress'] if on v0.10.4 or later, otherwise returns an
    empty list
    """
    if "chocolatey._no_progress" in __context__:
        return __context__["chocolatey._no_progress"]
    if Version(chocolatey_version()) >= Version("0.10.4"):
        answer = ["--no-progress"]
    else:
        log.warning("--no-progress unsupported in choco < 0.10.4")
        answer = []
    __context__["chocolatey._no_progress"] = answer
    return __context__["chocolatey._no_progress"]


def _find_chocolatey():
    """
    Returns the full path to chocolatey.bat on the host.
    """
    # Check context
    if "chocolatey._path" in __context__:
        return __context__["chocolatey._path"]

    # Check the path
    choc_path = __salt__["cmd.which"]("chocolatey.exe")
    if choc_path:
        __context__["chocolatey._path"] = choc_path
        return __context__["chocolatey._path"]

    # Check in common locations
    choc_defaults = [
        os.path.join(
            os.environ.get("ProgramData"), "Chocolatey", "bin", "chocolatey.exe"
        ),
        os.path.join(os.environ.get("ProgramData"), "Chocolatey", "bin", "choco.exe"),
        os.path.join(
            os.environ.get("SystemDrive"), "Chocolatey", "bin", "chocolatey.bat"
        ),
    ]
    for choc_exe in choc_defaults:
        if os.path.isfile(choc_exe):
            __context__["chocolatey._path"] = choc_exe
            return __context__["chocolatey._path"]

    # Not installed, raise an error
    err = (
        "Chocolatey not installed. Use chocolatey.bootstrap to "
        "install the Chocolatey package manager."
    )
    raise CommandExecutionError(err)


def chocolatey_version():
    """
    Returns the version of Chocolatey installed on the minion.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.chocolatey_version
    """
    if "chocolatey._version" in __context__:
        return __context__["chocolatey._version"]

    cmd = [_find_chocolatey()]
    cmd.append("-v")
    out = __salt__["cmd.run"](cmd, python_shell=False)
    __context__["chocolatey._version"] = out

    return __context__["chocolatey._version"]


def bootstrap(force=False, source=None):
    """
    Download and install the latest version of the Chocolatey package manager
    via the official bootstrap.

    Chocolatey requires Windows PowerShell and the .NET v4.0 runtime. Depending
    on the host's version of Windows, chocolatey.bootstrap will attempt to
    ensure these prerequisites are met by downloading and executing the
    appropriate installers from Microsoft.

    .. note::
        If PowerShell is installed, you may have to restart the host machine for
        Chocolatey to work.

    .. note::
        If you're installing offline using the source parameter, the PowerShell
        and .NET requirements must already be met on the target. This shouldn't
        be a problem on Windows versions 2012/8 and later

    Args:

        force (bool):
            Run the bootstrap process even if Chocolatey is found in the path.

        source (str):
            The location of the ``.nupkg`` file or ``.ps1`` file to run from an
            alternate location. This can be one of the following types of URLs:

            - salt://
            - http(s)://
            - ftp://
            - file:// - A local file on the system

            .. versionadded:: 3001

    Returns:
        str: The stdout of the Chocolatey installation script

    CLI Example:

    .. code-block:: bash

        # To bootstrap Chocolatey
        salt '*' chocolatey.bootstrap
        salt '*' chocolatey.bootstrap force=True

        # To bootstrap Chocolatey offline from a file on the salt master
        salt '*' chocolatey.bootstrap source=salt://files/chocolatey.nupkg

        # To bootstrap Chocolatey from a file on C:\\Temp
        salt '*' chocolatey.bootstrap source=C:\\Temp\\chocolatey.nupkg
    """
    # Check if Chocolatey is already present in the path
    try:
        choc_path = _find_chocolatey()
    except CommandExecutionError:
        choc_path = None
    if choc_path and not force:
        return f"Chocolatey found at {choc_path}"

    temp_dir = tempfile.gettempdir()

    # Make sure PowerShell is on the System if we're passing source
    # Vista and Windows Server 2008 do not have Powershell installed
    powershell_info = __salt__["cmd.shell_info"](shell="powershell")
    if not powershell_info["installed"]:
        # The following lookup tables are required to determine the correct
        # download required to install PowerShell. That's right, there's more
        # than one! You're welcome.
        ps_downloads = {
            (
                "Vista",
                "x86",
            ): "http://download.microsoft.com/download/A/7/5/A75BC017-63CE-47D6-8FA4-AFB5C21BAC54/Windows6.0-KB968930-x86.msu",
            (
                "Vista",
                "AMD64",
            ): "http://download.microsoft.com/download/3/C/8/3C8CF51E-1D9D-4DAA-AAEA-5C48D1CD055C/Windows6.0-KB968930-x64.msu",
            (
                "2008Server",
                "x86",
            ): "http://download.microsoft.com/download/F/9/E/F9EF6ACB-2BA8-4845-9C10-85FC4A69B207/Windows6.0-KB968930-x86.msu",
            (
                "2008Server",
                "AMD64",
            ): "http://download.microsoft.com/download/2/8/6/28686477-3242-4E96-9009-30B16BED89AF/Windows6.0-KB968930-x64.msu",
        }

        # PowerShell needs to be installed on older systems (Vista, 2008Server)
        if (__grains__["osrelease"], __grains__["cpuarch"]) in ps_downloads:

            # Install the appropriate release of PowerShell v2.0
            url = ps_downloads[(__grains__["osrelease"], __grains__["cpuarch"])]
            dest = os.path.join(temp_dir, os.path.basename(url))
            # Download the KB
            try:
                log.debug("Downloading PowerShell...")
                __salt__["cp.get_url"](path=url, dest=dest)
            except MinionError:
                err = "Failed to download PowerShell KB for {}".format(
                    __grains__["osrelease"]
                )
                if source:
                    raise CommandExecutionError(
                        "{}: PowerShell is required to bootstrap Chocolatey "
                        "with Source".format(err)
                    )
                raise CommandExecutionError(err)
            # Install the KB
            cmd = [dest, "/quiet", "/norestart"]
            log.debug("Installing PowerShell...")
            result = __salt__["cmd.run_all"](cmd, python_shell=False)
            if result["retcode"] != 0:
                err = (
                    "Failed to install PowerShell KB. For more information "
                    "run the installer manually on the host"
                )
                raise CommandExecutionError(err)
        else:
            err = "Windows PowerShell Installation not available"
            raise CommandNotFoundError(err)

    # Check that .NET v4.0+ is installed
    # Windows 7 / Windows Server 2008 R2 and below do not come with at least
    # .NET v4.0 installed
    if not __utils__["dotnet.version_at_least"](version="4"):
        # It took until .NET v4.0 for Microsoft got the hang of making
        # installers, this should work under any version of Windows
        url = "http://download.microsoft.com/download/1/B/E/1BE39E79-7E39-46A3-96FF-047F95396215/dotNetFx40_Full_setup.exe"
        dest = os.path.join(temp_dir, os.path.basename(url))
        # Download the .NET Framework 4 web installer
        try:
            log.debug("Downloading .NET v4.0...")
            __salt__["cp.get_url"](path=url, dest=dest)
        except MinionError:
            err = "Failed to download .NET v4.0 Web Installer"
            if source:
                err = (
                    "{}: .NET v4.0+ is required to bootstrap "
                    "Chocolatey with Source".format(err)
                )
            raise CommandExecutionError(err)

        # Run the .NET Framework 4 web installer
        cmd = [dest, "/q", "/norestart"]
        log.debug("Installing .NET v4.0...")
        result = __salt__["cmd.run_all"](cmd, python_shell=False)
        if result["retcode"] != 0:
            err = (
                "Failed to install .NET v4.0 failed. For more information "
                "run the installer manually on the host"
            )
            raise CommandExecutionError(err)

    # Define target / destination
    if source:
        url = source
    else:
        url = "https://chocolatey.org/install.ps1"
    dest = os.path.join(temp_dir, os.path.basename(url))

    # Download Chocolatey installer
    try:
        log.debug("Downloading Chocolatey: %s", os.path.basename(url))
        script = __salt__["cp.get_url"](path=url, dest=dest)
        log.debug("Script: %s", script)
    except MinionError:
        err = "Failed to download Chocolatey Installer"
        if source:
            err = "{0} from source"
        raise CommandExecutionError(err)

    # If this is a nupkg download we need to unzip it first
    if os.path.splitext(os.path.basename(dest))[1] == ".nupkg":
        log.debug("Unzipping Chocolatey: %s", dest)
        __salt__["archive.unzip"](
            zip_file=dest,
            dest=os.path.join(os.path.dirname(dest), "chocolatey"),
            extract_perms=False,
        )
        script = os.path.join(
            os.path.dirname(dest), "chocolatey", "tools", "chocolateyInstall.ps1"
        )

    if not os.path.exists(script):
        raise CommandExecutionError(
            f"Failed to find Chocolatey installation script: {script}"
        )

    # Run the Chocolatey bootstrap
    log.debug("Installing Chocolatey: %s", script)
    result = __salt__["cmd.script"](
        script, cwd=os.path.dirname(script), shell="powershell", python_shell=True
    )
    if result["retcode"] != 0:
        err = "Bootstrapping Chocolatey failed: {}".format(result["stderr"])
        raise CommandExecutionError(err)

    return result["stdout"]


def unbootstrap():
    """
    Uninstall chocolatey from the system by doing the following:

    - Delete the Chocolatey Directory
    - Remove Chocolatey from the path
    - Remove Chocolatey environment variables

    .. versionadded:: 3001

    Returns:
        list: A list of items that were removed, otherwise an empty list

    CLI Example:

    .. code-block:: bash

        salt * chocolatey.unbootstrap
    """
    removed = []

    # Delete the Chocolatey directory
    choco_dir = os.environ.get("ChocolateyInstall", False)
    if choco_dir:
        if os.path.exists(choco_dir):
            log.debug("Removing Chocolatey directory: %s", choco_dir)
            __salt__["file.remove"](path=choco_dir, force=True)
            removed.append(f"Removed Directory: {choco_dir}")
    else:
        known_paths = [
            os.path.join(os.environ.get("ProgramData"), "Chocolatey"),
            os.path.join(os.environ.get("SystemDrive"), "Chocolatey"),
        ]
        for path in known_paths:
            if os.path.exists(path):
                log.debug("Removing Chocolatey directory: %s", path)
                __salt__["file.remove"](path=path, force=True)
                removed.append(f"Removed Directory: {path}")

    # Delete all Chocolatey environment variables
    for env_var in __salt__["environ.items"]():
        if env_var.lower().startswith("chocolatey"):
            log.debug("Removing Chocolatey environment variable: %s", env_var)
            __salt__["environ.setval"](
                key=env_var, val=False, false_unsets=True, permanent="HKLM"
            )
            __salt__["environ.setval"](
                key=env_var, val=False, false_unsets=True, permanent="HKCU"
            )
            removed.append(f"Removed Environment Var: {env_var}")

    # Remove Chocolatey from the path:
    for path in __salt__["win_path.get_path"]():
        if "chocolatey" in path.lower():
            log.debug("Removing Chocolatey path item: %s", path)
            __salt__["win_path.remove"](path=path, rehash=True)
            removed.append(f"Removed Path Item: {path}")

    return removed


def list_(
    narrow=None,
    all_versions=False,
    pre_versions=False,
    source=None,
    local_only=False,
    exact=False,
):
    """
    Instructs Chocolatey to pull a vague package list from the repository.

    Args:

        narrow (str):
            Term used to narrow down results. Searches against
            name/description/tag. Default is None.

        all_versions (bool):
            Display all available package versions in results. Default is False.

        pre_versions (bool):
            Display pre-release packages in results. Default is False.

        source (str):
            Chocolatey repository (directory, share or remote URL feed) the
            package comes from. Defaults to the official Chocolatey feed if
            None is passed. Default is None.

        local_only (bool):
            Only display packages that are installed locally. Default is False.

        exact (bool):
            Only display packages that match ``narrow`` exactly. Default is
            False.

            .. versionadded:: 2017.7.0

    Returns:
        dict: A dictionary of results.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.list <narrow>
        salt '*' chocolatey.list <narrow> all_versions=True
    """
    choc_path = _find_chocolatey()
    # https://docs.chocolatey.org/en-us/guides/upgrading-to-chocolatey-v2-v6
    if Version(chocolatey_version()) < Version("2.0.0"):
        cmd = [choc_path, "list"]
        if local_only:
            cmd.append("--local-only")
    else:
        if local_only:
            # Starting with 2.0.0, list only returns local packages
            cmd = [choc_path, "list"]
        else:
            cmd = [choc_path, "search"]
    if narrow:
        cmd.append(narrow)
    if salt.utils.data.is_true(all_versions):
        cmd.append("--allversions")
    if salt.utils.data.is_true(pre_versions):
        cmd.append("--prerelease")
    if source:
        cmd.extend(["--source", source])
    if exact:
        cmd.append("--exact")

    # This is needed to parse the output correctly
    cmd.append("--limit-output")

    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    # Chocolatey introduced Enhanced Exit Codes starting with version 0.10.12
    # Exit Code 2 means there were no results, but is not a failure
    # This may start to effect other functions in the future as Chocolatey
    # moves more functions to this new paradigm
    # https://github.com/chocolatey/choco/issues/1758
    if result["retcode"] not in [0, 2]:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    ret = CaseInsensitiveDict({})
    pkg_re = re.compile(r"(\S+)\|(\S+)")
    for line in result["stdout"].split("\n"):
        if line.startswith("No packages"):
            return ret
        for name, ver in pkg_re.findall(line):
            if "chocolatey" in name:
                continue
            if name not in ret:
                ret[name] = []
            ret[name].append(ver)

    return ret


def list_webpi():
    """
    Instructs Chocolatey to pull a full package list from the Microsoft Web PI
    repository.

    Returns:
        str: List of webpi packages

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.list_webpi
    """
    choc_path = _find_chocolatey()
    # https://docs.chocolatey.org/en-us/guides/upgrading-to-chocolatey-v2-v6
    if Version(chocolatey_version()) < Version("2.0.0"):
        cmd = [choc_path, "list", "--source", "webpi"]
    else:
        cmd = [choc_path, "search", "--source", "webpi"]
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] != 0:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    return result["stdout"]


def list_windowsfeatures():
    """
    Instructs Chocolatey to pull a full package list from the Windows Features
    list, via the Deployment Image Servicing and Management tool.

    Returns:
        str: List of Windows Features

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.list_windowsfeatures
    """
    choc_path = _find_chocolatey()
    # https://docs.chocolatey.org/en-us/guides/upgrading-to-chocolatey-v2-v6
    if Version(chocolatey_version()) < Version("2.0.0"):
        cmd = [choc_path, "list", "--source", "windowsfeatures"]
    else:
        cmd = [choc_path, "search", "--source", "windowsfeatures"]
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] != 0:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    return result["stdout"]


def install(
    name,
    version=None,
    source=None,
    force=False,
    pre_versions=False,
    install_args=None,
    override_args=False,
    force_x86=False,
    package_args=None,
    allow_multiple=False,
    execution_timeout=None,
):
    """
    Instructs Chocolatey to install a package.

    Args:

        name (str):
            The name of the package to be installed. Only accepts a single
            argument. Required.

        version (str):
            Install a specific version of the package. Defaults to latest
            version. Default is ``None``.

        source (str):
            Chocolatey repository (directory, share or remote URL feed) the
            package comes from. Defaults to the official Chocolatey feed.
            Default is ``None``.

            Alternate Sources:

            - cygwin
            - python
            - ruby
            - webpi
            - windowsfeatures

        force (bool):
            Reinstall the current version of an existing package. Do not use
            with ``allow_multiple``. Default is ``False``.

        pre_versions (bool):
            Include pre-release packages. Default is ``False``.

        install_args (str):
            A list of install arguments you want to pass to the installation
            process, i.e. product key or feature list. Default is ``None``.

        override_args (bool):
            Set to true if you want to override the original install arguments
            (for the native installer) in the package and use your own. When
            this is set to ``False`` install_args will be appended to the end of
            the default arguments. Default is ``None``.

        force_x86 (bool):
            Force x86 (32bit) installation on 64bit systems. Default is
            ``False``.

        package_args (str):
            Arguments you want to pass to the package. Default is ``None``.

        allow_multiple (bool):
            Allow multiple versions of the package to be installed. Do not use
            with ``force``. Does not work with all packages. Default is
            ``False``.

            .. versionadded:: 2017.7.0

        execution_timeout (str):
            Chocolatey execution timeout value you want to pass to the
            installation process. Default is ``None``.

            .. versionadded:: 2018.3.0

    Returns:
        str: The output of the ``chocolatey`` command

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.install <package name>
        salt '*' chocolatey.install <package name> version=<package version>
        salt '*' chocolatey.install <package name> install_args=<args> override_args=True
    """
    if force and allow_multiple:
        raise SaltInvocationError(
            "Cannot use 'force' in conjunction with 'allow_multiple'"
        )

    choc_path = _find_chocolatey()
    # chocolatey helpfully only supports a single package argument
    # CORRECTION: it also supports multiple package names separated by spaces
    # but any additional arguments apply to ALL packages specified
    cmd = [choc_path, "install", name]
    if version:
        cmd.extend(["--version", version])
    if source:
        cmd.extend(["--source", source])
    if salt.utils.data.is_true(force):
        cmd.append("--force")
    if salt.utils.data.is_true(pre_versions):
        cmd.append("--prerelease")
    if install_args:
        cmd.extend(["--installarguments", install_args])
    if override_args:
        cmd.append("--overridearguments")
    if force_x86:
        cmd.append("--forcex86")
    if package_args:
        cmd.extend(["--packageparameters", package_args])
    if allow_multiple:
        cmd.append("--allow-multiple")
    if execution_timeout:
        cmd.extend(["--execution-timeout", execution_timeout])

    # Salt doesn't need to see the progress
    cmd.extend(_no_progress())
    cmd.extend(_yes())
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] not in [0, 1641, 3010]:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    if name == "chocolatey":
        _clear_context()

    return result["stdout"]


def install_cygwin(name, install_args=None, override_args=False):
    """
    Instructs Chocolatey to install a package via Cygwin.

    Args:

        name (str):
            The name of the package to be installed. Only accepts a single
            argument.

        install_args (str):
            A list of install arguments you want to pass to the installation
            process, i.e. product key or feature list

        override_args (bool):
            Set to ``True`` if you want to override the original install
            arguments (for the native installer) in the package and use your
            own. When this is set to ``False`` install_args will be appended to
            the end of the default arguments

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.install_cygwin <package name>
        salt '*' chocolatey.install_cygwin <package name> install_args=<args> override_args=True
    """
    return install(
        name, source="cygwin", install_args=install_args, override_args=override_args
    )


def install_gem(name, version=None, install_args=None, override_args=False):
    """
    Instructs Chocolatey to install a package via Ruby's Gems.

    Args:

        name (str):
            The name of the package to be installed. Only accepts a single
            argument.

        version (str):
            Install a specific version of the package. Defaults to the latest
            version available.

        install_args (str):
            A list of install arguments you want to pass to the installation
            process, i.e. product key or feature list

        override_args (bool):
            Set to ``True`` if you want to override the original install
            arguments (for the native installer) in the package and use your
            own. When this is set to ``False`` install_args will be appended to
            the end of the default arguments

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.install_gem <package name>
        salt '*' chocolatey.install_gem <package name> version=<package version>
        salt '*' chocolatey.install_gem <package name> install_args=<args> override_args=True
    """
    return install(
        name,
        version=version,
        source="ruby",
        install_args=install_args,
        override_args=override_args,
    )


def install_missing(name, version=None, source=None):
    """
    Instructs Chocolatey to install a package if it doesn't already exist.

    .. versionchanged:: 2014.7.0
        If the minion has Chocolatey >= 0.9.8.24 installed, this function calls
        :mod:`chocolatey.install <salt.modules.chocolatey.install>` instead, as
        ``installmissing`` is deprecated as of that version and will be removed
        in Chocolatey 1.0.

    Args:

        name (str):
            The name of the package to be installed. Only accepts a single
            argument.

        version (str):
            Install a specific version of the package. Defaults to the latest
            version available.

        source (str):
            Chocolatey repository (directory, share or remote URL feed) the
            package comes from. Defaults to the official Chocolatey feed.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.install_missing <package name>
        salt '*' chocolatey.install_missing <package name> version=<package version>
    """
    if Version(chocolatey_version()) >= Version("0.9.8.24"):
        log.warning("installmissing is deprecated, using install")
        return install(name, version=version)

    # chocolatey helpfully only supports a single package argument
    cmd = [_find_chocolatey(), "installmissing", name]
    if version:
        cmd.extend(["--version", version])
    if source:
        cmd.extend(["--source", source])
    # Shouldn't need this as this code should never run on v0.9.9 and newer
    cmd.extend(_yes())
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] != 0:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    return result["stdout"]


def install_python(name, version=None, install_args=None, override_args=False):
    """
    Instructs Chocolatey to install a package via Python's easy_install.

    Args:

        name (str):
            The name of the package to be installed. Only accepts a single
            argument.

        version (str):
            Install a specific version of the package. Defaults to the latest
            version available.

        install_args (str):
            A list of install arguments you want to pass to the installation
            process, i.e. product key or feature list.

        override_args (bool):
            Set to ``True`` if you want to override the original install
            arguments (for the native installer) in the package and use your
            own. When this is set to ``False`` install_args will be appended to
            the end of the default arguments.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.install_python <package name>
        salt '*' chocolatey.install_python <package name> version=<package version>
        salt '*' chocolatey.install_python <package name> install_args=<args> override_args=True
    """
    return install(
        name,
        version=version,
        source="python",
        install_args=install_args,
        override_args=override_args,
    )


def install_windowsfeatures(name):
    """
    Instructs Chocolatey to install a Windows Feature via the Deployment Image
    Servicing and Management tool.

    Args:

        name (str):
            The name of the feature to be installed. Only accepts a single
            argument.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.install_windowsfeatures <package name>
    """
    return install(name, source="windowsfeatures")


def install_webpi(name, install_args=None, override_args=False):
    """
    Instructs Chocolatey to install a package via the Microsoft Web PI service.

    Args:

        name (str):
            The name of the package to be installed. Only accepts a single
            argument.

        install_args (str):
            A list of install arguments you want to pass to the installation
            process, i.e. product key or feature list.

        override_args (bool):
            Set to ``True`` if you want to override the original install
            arguments (for the native installer) in the package and use your
            own. When this is set to ``False`` install_args will be appended to
            the end of the default arguments.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.install_webpi <package name>
        salt '*' chocolatey.install_webpi <package name> install_args=<args> override_args=True
    """
    return install(
        name, source="webpi", install_args=install_args, override_args=override_args
    )


def uninstall(
    name,
    version=None,
    uninstall_args=None,
    override_args=False,
    force=False,
):
    """
    Instructs Chocolatey to uninstall a package.

    Args:

        name (str):
            The name of the package to be uninstalled. Only accepts a single
            argument.

        version (str):
            Uninstalls a specific version of the package. Defaults to the latest
            version installed.

        uninstall_args (str):
            A list of uninstall arguments you want to pass to the uninstallation
            process, i.e. product key or feature list.

        override_args
            Set to ``True`` if you want to override the original uninstall
            arguments (for the native uninstaller) in the package and use your
            own. When this is set to ``False`` uninstall_args will be appended
            to the end of the default arguments.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.uninstall <package name>
        salt '*' chocolatey.uninstall <package name> version=<package version>
        salt '*' chocolatey.uninstall <package name> version=<package version> uninstall_args=<args> override_args=True
    """
    # chocolatey helpfully only supports a single package argument
    cmd = [_find_chocolatey(), "uninstall", name]
    if version:
        cmd.extend(["--version", version])
    if uninstall_args:
        cmd.extend(["--uninstallarguments", uninstall_args])
    if override_args:
        cmd.append("--overridearguments")
    if force:
        cmd.append("--force")
    cmd.extend(_yes())
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] not in [0, 1, 1605, 1614, 1641]:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    return result["stdout"]


def upgrade(
    name,
    version=None,
    source=None,
    force=False,
    pre_versions=False,
    install_args=None,
    override_args=False,
    force_x86=False,
    package_args=None,
):
    """
    .. versionadded:: 2016.3.4

    Instructs Chocolatey to upgrade packages on the system. (update is being
    deprecated). This command will install the package if not installed.

    Args:

        name (str):
            The name of the package to update, or "all" to update everything
            installed on the system.

        version (str):
            Install a specific version of the package. Defaults to latest
            version.

        source (str):
            Chocolatey repository (directory, share or remote URL feed) the
            package comes from. Defaults to the official Chocolatey feed.

        force (bool):
            Reinstall the **same** version already installed.

        pre_versions (bool):
            Include pre-release packages in comparison. Defaults to ``False``.

        install_args (str):
            A list of install arguments you want to pass to the installation
            process, i.e. product key or feature list.

        override_args (bool):
            Set to ``True`` if you want to override the original install
            arguments (for the native installer) in the package and use your
            own. When this is set to ``False`` install_args will be appended to
            the end of the default arguments.

        force_x86 (bool):
            Force x86 (32bit) installation on 64bit systems. Defaults to
            ``False``.

        package_args (str):
            A list of arguments you want to pass to the package.

    Returns:
        str: Results of the ``chocolatey`` command

    CLI Example:

    .. code-block:: bash

        salt "*" chocolatey.upgrade all
        salt "*" chocolatey.upgrade <package name> pre_versions=True
    """
    # chocolatey helpfully only supports a single package argument
    cmd = [_find_chocolatey(), "upgrade", name]
    if version:
        cmd.extend(["--version", version])
    if source:
        cmd.extend(["--source", source])
    if salt.utils.data.is_true(force):
        cmd.append("--force")
    if salt.utils.data.is_true(pre_versions):
        cmd.append("--prerelease")
    if install_args:
        cmd.extend(["--installarguments", install_args])
    if override_args:
        cmd.append("--overridearguments")
    if force_x86:
        cmd.append("--forcex86")
    if package_args:
        cmd.extend(["--packageparameters", package_args])

    # Salt doesn't need to see the progress
    cmd.extend(_no_progress())
    cmd.extend(_yes())

    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] not in [0, 1641, 3010]:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    return result["stdout"]


def update(name, source=None, pre_versions=False):
    """
    Instructs Chocolatey to update packages on the system.

    Args:

        name (str):
            The name of the package to update, or "all" to update everything
            installed on the system.

        source (str):
            Chocolatey repository (directory, share or remote URL feed) the
            package comes from. Defaults to the official Chocolatey feed.

        pre_versions (bool):
            Include pre-release packages in comparison. Defaults to ``False``.

    CLI Example:

    .. code-block:: bash

        salt "*" chocolatey.update all
        salt "*" chocolatey.update <package name> pre_versions=True
    """
    # chocolatey helpfully only supports a single package argument
    if Version(chocolatey_version()) >= Version("0.9.8.24"):
        log.warning("update is deprecated, using upgrade")
        return upgrade(name, source=source, pre_versions=pre_versions)

    cmd = [_find_chocolatey(), "update", name]
    if source:
        cmd.extend(["--source", source])
    if salt.utils.data.is_true(pre_versions):
        cmd.append("--prerelease")

    # Salt doesn't need to see the progress
    cmd.extend(_no_progress())
    cmd.extend(_yes())
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] not in [0, 1641, 3010]:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    return result["stdout"]


def version(name, check_remote=False, source=None, pre_versions=False):
    """
    Instructs Chocolatey to check an installed package version, and optionally
    compare it to one available from a remote feed.

    Args:

        name (str):
            The name of the package to check. Required.

        check_remote (bool):
            Get the version number of the latest package from the remote feed.
            Default is ``False``.

        source (str):
            Chocolatey repository (directory, share or remote URL feed) the
            package comes from. Defaults to the official Chocolatey feed.
            Default is ``None``.

        pre_versions (bool):
            Include pre-release packages in comparison. Default is ``False``.

    Returns:
        dict: A dictionary of currently installed software and versions

    CLI Example:

    .. code-block:: bash

        salt "*" chocolatey.version <package name>
        salt "*" chocolatey.version <package name> check_remote=True
    """
    installed = list_(narrow=name, local_only=True)

    packages = {}
    lower_name = name.lower()
    if installed:
        for pkg in installed:
            if lower_name == pkg.lower():
                packages.setdefault(lower_name, {})
                packages[lower_name]["installed"] = installed[pkg]

    if check_remote:
        # If there's a remote package available, then also include that
        # in the dictionary that we return.
        available = list_(
            narrow=name, local_only=False, pre_versions=pre_versions, source=source
        )
        if available:
            for pkg in available:
                if lower_name == pkg.lower():
                    packages.setdefault(lower_name, {})
                    packages[lower_name]["available"] = available[pkg]

    return packages


def add_source(name, source_location, username=None, password=None, priority=None):
    """
    Instructs Chocolatey to add a source.

    Args:

        name (str):
            The name of the source to be added as a chocolatey repository.

        source (str):
            Location of the source you want to work with.

        username (str):
            Provide username for chocolatey sources that need authentication
            credentials.

        password (str):
            Provide password for chocolatey sources that need authentication
            credentials.

        priority (int):
            The priority order of this source as compared to other sources,
            lower is better. Defaults to 0 (no priority). All priorities
            above 0 will be evaluated first, then zero-based values will be
            evaluated in config file order.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.add_source <source name> <source_location>
        salt '*' chocolatey.add_source <source name> <source_location> priority=100
        salt '*' chocolatey.add_source <source name> <source_location> user=<user> password=<password>

    """
    cmd = [
        _find_chocolatey(),
        "sources",
        "add",
        "--name",
        name,
        "--source",
        source_location,
    ]
    if username:
        cmd.extend(["--user", username])
    if password:
        cmd.extend(["--password", password])
    if priority:
        cmd.extend(["--priority", priority])
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] != 0:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    return result["stdout"]


def _change_source_state(name, state):
    """
    Instructs Chocolatey to change the state of a source.

    Args:

        name (str):
            Name of the repository to affect.

        state (str):
            State in which you want the chocolatey repository.
    """
    cmd = [_find_chocolatey(), "source", state, "--name", name]
    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if result["retcode"] != 0:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    return result["stdout"]


def enable_source(name):
    """
    Instructs Chocolatey to enable a source.

    Args:

        name (str):
            Name of the source repository to enable.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.enable_source <name>

    """
    return _change_source_state(name, "enable")


def disable_source(name):
    """
    Instructs Chocolatey to disable a source.

    Args:

        name (str):
            Name of the source repository to disable.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.disable_source <name>
    """
    return _change_source_state(name, "disable")


def list_sources():
    """
    Returns the list of installed sources.

    CLI Example:

    .. code-block:: bash

        salt '*' chocolatey.list_sources
    """
    choc_path = _find_chocolatey()
    cmd = [choc_path, "source"]

    # This is needed to parse the output correctly
    cmd.append("--limit-output")

    result = __salt__["cmd.run_all"](cmd, python_shell=False)

    # Chocolatey introduced Enhanced Exit Codes starting with version 0.10.12
    # Exit Code 2 means there were no results, but is not a failure
    # This may start to effect other functions in the future as Chocolatey
    # moves more functions to this new paradigm
    # https://github.com/chocolatey/choco/issues/1758
    if result["retcode"] not in [0, 2]:
        err = "Running chocolatey failed: {}".format(result["stdout"])
        raise CommandExecutionError(err)

    ret = CaseInsensitiveDict({})
    pkg_re = re.compile(r"(.*)\|(.*)\|(.*)\|(.*)\|.*\|.*\|.*\|.*\|.*")
    for line in result["stdout"].split("\n"):
        for name, url, disabled, user in pkg_re.findall(line):
            if name not in ret:
                ret[name] = {"URL: ": url, "Disabled": disabled, "User: ": user}

    return ret



"""
Support for Apache

Please note: The functions in here are Debian-specific. Placing them in this
separate file will allow them to load only on Debian-based systems, while still
loading under the ``apache`` namespace.
"""

import logging
import os

import salt.utils.decorators.path
import salt.utils.path

log = logging.getLogger(__name__)

__virtualname__ = "apache"

SITE_ENABLED_DIR = "/etc/apache2/sites-enabled"


def __virtual__():
    """
    Only load the module if apache is installed
    """
    cmd = _detect_os()
    if salt.utils.path.which(cmd) and __grains__["os_family"] == "Debian":
        return __virtualname__
    return (False, "apache execution module not loaded: apache not installed.")


def _detect_os():
    """
    Apache commands and paths differ depending on packaging
    """
    # TODO: Add pillar support for the apachectl location
    if __grains__["os_family"] == "RedHat":
        return "apachectl"
    elif __grains__["os_family"] == "Debian":
        return "apache2ctl"
    else:
        return "apachectl"


def check_site_enabled(site):
    """
    Checks to see if the specific site symlink is in /etc/apache2/sites-enabled.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.check_site_enabled example.com
        salt '*' apache.check_site_enabled example.com.conf
    """
    if site.endswith(".conf"):
        site_file = site
    else:
        site_file = f"{site}.conf"
    if os.path.islink(f"{SITE_ENABLED_DIR}/{site_file}"):
        return True
    elif site == "default" and os.path.islink(f"{SITE_ENABLED_DIR}/000-{site_file}"):
        return True
    else:
        return False


def a2ensite(site):
    """
    Runs a2ensite for the given site.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.a2ensite example.com
    """
    ret = {}
    command = ["a2ensite", site]

    try:
        status = __salt__["cmd.retcode"](command, python_shell=False)
    except Exception as e:  # pylint: disable=broad-except
        return e

    ret["Name"] = "Apache2 Enable Site"
    ret["Site"] = site

    if status == 1:
        ret["Status"] = f"Site {site} Not found"
    elif status == 0:
        ret["Status"] = f"Site {site} enabled"
    else:
        ret["Status"] = status

    return ret


def a2dissite(site):
    """
    Runs a2dissite for the given site.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.a2dissite example.com
    """
    ret = {}
    command = ["a2dissite", site]

    try:
        status = __salt__["cmd.retcode"](command, python_shell=False)
    except Exception as e:  # pylint: disable=broad-except
        return e

    ret["Name"] = "Apache2 Disable Site"
    ret["Site"] = site

    if status == 256:
        ret["Status"] = f"Site {site} Not found"
    elif status == 0:
        ret["Status"] = f"Site {site} disabled"
    else:
        ret["Status"] = status

    return ret


def check_mod_enabled(mod):
    """
    Checks to see if the specific mod symlink is in /etc/apache2/mods-enabled.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.check_mod_enabled status
        salt '*' apache.check_mod_enabled status.load
        salt '*' apache.check_mod_enabled status.conf
    """
    if mod.endswith(".load") or mod.endswith(".conf"):
        mod_file = mod
    else:
        mod_file = f"{mod}.load"
    return os.path.islink(f"/etc/apache2/mods-enabled/{mod_file}")


def a2enmod(mod):
    """
    Runs a2enmod for the given mod.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.a2enmod vhost_alias
    """
    ret = {}
    command = ["a2enmod", mod]

    try:
        status = __salt__["cmd.retcode"](command, python_shell=False)
    except Exception as e:  # pylint: disable=broad-except
        return e

    ret["Name"] = "Apache2 Enable Mod"
    ret["Mod"] = mod

    if status == 1:
        ret["Status"] = f"Mod {mod} Not found"
    elif status == 0:
        ret["Status"] = f"Mod {mod} enabled"
    else:
        ret["Status"] = status

    return ret


def a2dismod(mod):
    """
    Runs a2dismod for the given mod.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.a2dismod vhost_alias
    """
    ret = {}
    command = ["a2dismod", mod]

    try:
        status = __salt__["cmd.retcode"](command, python_shell=False)
    except Exception as e:  # pylint: disable=broad-except
        return e

    ret["Name"] = "Apache2 Disable Mod"
    ret["Mod"] = mod

    if status == 256:
        ret["Status"] = f"Mod {mod} Not found"
    elif status == 0:
        ret["Status"] = f"Mod {mod} disabled"
    else:
        ret["Status"] = status

    return ret


def check_conf_enabled(conf):
    """
    .. versionadded:: 2016.3.0

    Checks to see if the specific conf symlink is in /etc/apache2/conf-enabled.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.check_conf_enabled security
        salt '*' apache.check_conf_enabled security.conf
    """
    if conf.endswith(".conf"):
        conf_file = conf
    else:
        conf_file = f"{conf}.conf"
    return os.path.islink(f"/etc/apache2/conf-enabled/{conf_file}")


@salt.utils.decorators.path.which("a2enconf")
def a2enconf(conf):
    """
    .. versionadded:: 2016.3.0

    Runs a2enconf for the given conf.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.a2enconf security
    """
    ret = {}
    command = ["a2enconf", conf]

    try:
        status = __salt__["cmd.retcode"](command, python_shell=False)
    except Exception as e:  # pylint: disable=broad-except
        return e

    ret["Name"] = "Apache2 Enable Conf"
    ret["Conf"] = conf

    if status == 1:
        ret["Status"] = f"Conf {conf} Not found"
    elif status == 0:
        ret["Status"] = f"Conf {conf} enabled"
    else:
        ret["Status"] = status

    return ret


@salt.utils.decorators.path.which("a2disconf")
def a2disconf(conf):
    """
    .. versionadded:: 2016.3.0

    Runs a2disconf for the given conf.

    This will only be functional on Debian-based operating systems (Ubuntu,
    Mint, etc).

    CLI Examples:

    .. code-block:: bash

        salt '*' apache.a2disconf security
    """
    ret = {}
    command = ["a2disconf", conf]

    try:
        status = __salt__["cmd.retcode"](command, python_shell=False)
    except Exception as e:  # pylint: disable=broad-except
        return e

    ret["Name"] = "Apache2 Disable Conf"
    ret["Conf"] = conf

    if status == 256:
        ret["Status"] = f"Conf {conf} Not found"
    elif status == 0:
        ret["Status"] = f"Conf {conf} disabled"
    else:
        ret["Status"] = status

    return ret



import pytest

import salt.utils.win_dacl as win_dacl

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def temp_path(tmp_path):
    # We need to create a directory that doesn't inherit permissions from the test suite
    tmp_path.mkdir(parents=True, exist_ok=True)
    win_dacl.set_owner(obj_name=str(tmp_path), principal="Administrators")
    assert win_dacl.get_owner(obj_name=str(tmp_path)) == "Administrators"
    # We don't want the parent test directory to inherit permissions
    win_dacl.set_inheritance(obj_name=str(tmp_path), enabled=False)
    assert not win_dacl.get_inheritance(obj_name=str(tmp_path))
    # Set these permissions and make sure they're the only ones
    win_dacl.set_permissions(
        obj_name=str(tmp_path),
        principal="Administrators",
        permissions="full_control",
        access_mode="grant",
        reset_perms=True,
        protected=True,
    )
    perms = {
        "Inherited": {},
        "Not Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            }
        },
    }
    assert win_dacl.get_permissions(obj_name=str(tmp_path)) == perms

    # Now we create a directory for testing that does inherit those permissions from the above, new parent directory
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()
    # We do want the test directory to inherit permissions from the parent directory
    assert win_dacl.get_inheritance(obj_name=str(test_dir))
    # Make sure the permissions are inherited from the parent
    perms = {
        "Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            }
        },
        "Not Inherited": {},
    }
    assert win_dacl.get_permissions(obj_name=str(test_dir)) == perms
    yield test_dir


def test_directory_new(file, temp_path):
    """
    Test file.directory when the directory does not exist
    Should just return "New Dir"
    """
    path = str(temp_path / "test")
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Administrators": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": "full_control"}},
    )
    expected = {path: {"directory": "new"}}
    assert ret["changes"] == expected
    permissions = win_dacl.get_permissions(path)
    expected = {
        "Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
        "Not Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
            "Guest": {
                "deny": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
    }
    assert permissions == expected


def test_directory_new_no_inherit(file, temp_path):
    """
    Test file.directory when the directory does not exist
    Should just return "New Dir"
    """
    path = str(temp_path / "test")
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Administrators": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": "full_control"}},
        win_inheritance=False,
    )
    expected = {path: {"directory": "new"}}
    assert ret["changes"] == expected
    assert not win_dacl.get_inheritance(path)
    permissions = win_dacl.get_permissions(path)
    assert permissions["Inherited"] == {}


def test_directory_new_reset(file, temp_path):
    """
    Test file.directory when the directory does not exist
    Should just return "New Dir"
    """
    path = str(temp_path / "test")
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Administrators": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": "full_control"}},
        win_perms_reset=True,
    )
    expected = {path: {"directory": "new"}}
    assert ret["changes"] == expected
    permissions = win_dacl.get_permissions(path)
    expected = {
        "Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
        "Not Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
            "Guest": {
                "deny": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
    }
    assert permissions == expected


def test_directory_new_reset_no_inherit(file, temp_path):
    """
    Test file.directory when the directory does not exist
    Should just return "New Dir"
    """
    path = str(temp_path / "test")
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Administrators": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": "full_control"}},
        win_inheritance=False,
        win_perms_reset=True,
    )
    expected = {path: {"directory": "new"}}
    assert ret["changes"] == expected
    permissions = win_dacl.get_permissions(path)
    expected = {
        "Inherited": {},
        "Not Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
            "Guest": {
                "deny": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
    }
    assert permissions == expected


def test_directory_existing(file, temp_path):
    path = str(temp_path)
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Everyone": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": ["write_data", "write_attributes"]}},
    )
    expected = {
        "deny_perms": {"Guest": {"permissions": ["write_data", "write_attributes"]}},
        "grant_perms": {"Everyone": {"permissions": "full_control"}},
    }
    # Sometimes an owner will be set, we don't care about the owner
    ret["changes"].pop("owner", None)
    assert ret["changes"] == expected
    permissions = win_dacl.get_permissions(path)
    expected = {
        "Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
        "Not Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
            "Everyone": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
            "Guest": {
                "deny": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": ["Create files / write data", "Write attributes"],
                }
            },
        },
    }
    assert permissions == expected


def test_directory_existing_existing_user(file, temp_path):
    path = str(temp_path)
    win_dacl.set_permissions(
        obj_name=path,
        principal="Everyone",
        permissions=["write_data", "write_attributes"],
        access_mode="grant",
        reset_perms=True,
    )
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Everyone": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": ["write_data", "write_attributes"]}},
    )
    expected = {
        "deny_perms": {"Guest": {"permissions": ["write_data", "write_attributes"]}},
        "grant_perms": {"Everyone": {"permissions": "full_control"}},
    }
    # Sometimes an owner will be set, we don't care about the owner
    ret["changes"].pop("owner", None)
    assert ret["changes"] == expected
    permissions = win_dacl.get_permissions(path)
    expected = {
        "Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
        "Not Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
            "Everyone": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
            "Guest": {
                "deny": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": ["Create files / write data", "Write attributes"],
                }
            },
        },
    }
    assert permissions == expected


def test_directory_existing_no_inherit(file, temp_path):
    path = str(temp_path)
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Everyone": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": ["write_data", "write_attributes"]}},
        win_inheritance=False,
    )
    expected = {
        "deny_perms": {"Guest": {"permissions": ["write_data", "write_attributes"]}},
        "grant_perms": {"Everyone": {"permissions": "full_control"}},
        "inheritance": False,
    }
    # Sometimes an owner will be set, we don't care about the owner
    ret["changes"].pop("owner", None)
    assert ret["changes"] == expected
    assert not win_dacl.get_inheritance(path)
    permissions = win_dacl.get_permissions(path)
    assert permissions["Inherited"] == {}


def test_directory_existing_reset(file, temp_path):
    path = str(temp_path)
    win_dacl.set_permissions(
        obj_name=path,
        principal="Guest",
        permissions=["write_data", "write_attributes"],
        access_mode="deny",
        reset_perms=True,
    )
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Everyone": {"perms": "full_control"}},
        win_perms_reset=True,
    )
    expected = {
        "grant_perms": {"Everyone": {"permissions": "full_control"}},
        "remove_perms": {
            "Guest": {
                "deny": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": ["Create files / write data", "Write attributes"],
                }
            }
        },
    }
    # Sometimes an owner will be set, we don't care about the owner
    ret["changes"].pop("owner", None)
    assert ret["changes"] == expected
    permissions = win_dacl.get_permissions(path)
    expected = {
        "Inherited": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
        "Not Inherited": {
            "Everyone": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
        },
    }
    assert permissions == expected


def test_directory_existing_reset_no_inherit(file, temp_path):
    path = str(temp_path)
    ret = file.directory(
        name=path,
        makedirs=True,
        win_perms={"Everyone": {"perms": "full_control"}},
        win_deny_perms={"Guest": {"perms": ["write_data", "write_attributes"]}},
        win_perms_reset=True,
        win_inheritance=False,
    )

    expected = {
        "deny_perms": {"Guest": {"permissions": ["write_data", "write_attributes"]}},
        "grant_perms": {"Everyone": {"permissions": "full_control"}},
        "inheritance": False,
        "remove_perms": {
            "Administrators": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                },
            },
        },
    }
    # Sometimes an owner will be set, we don't care about the owner
    ret["changes"].pop("owner", None)
    assert ret["changes"] == expected

    permissions = win_dacl.get_permissions(path)
    expected = {
        "Inherited": {},
        "Not Inherited": {
            "Everyone": {
                "grant": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": "Full control",
                }
            },
            "Guest": {
                "deny": {
                    "applies to": "This folder, subfolders and files",
                    "permissions": ["Create files / write data", "Write attributes"],
                }
            },
        },
    }
    assert permissions == expected



"""
Module for sending messages to Mattermost

.. versionadded:: 2017.7.0

:configuration: This module can be used by either passing an api_url and hook
    directly or by specifying both in a configuration profile in the salt
    master/minion config. For example:

    .. code-block:: yaml

        mattermost:
          hook: peWcBiMOS9HrZG15peWcBiMOS9HrZG15
          api_url: https://example.com
"""

import logging

import salt.utils.json

# pylint: disable=import-error,no-name-in-module,redefined-builtin
import salt.utils.mattermost

# pylint: enable=import-error,no-name-in-module
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)

__virtualname__ = "mattermost"


def __virtual__():
    """
    Return virtual name of the module.
    :return: The virtual name of the module.
    """
    return __virtualname__


def _get_hook():
    """
    Retrieves and return the Mattermost's configured hook
    :return:            String: the hook string
    """
    hook = __salt__["config.get"]("mattermost.hook") or __salt__["config.get"](
        "mattermost:hook"
    )
    if not hook:
        raise SaltInvocationError("No Mattermost Hook found")

    return hook


def _get_api_url():
    """
    Retrieves and return the Mattermost's configured api url
    :return:            String: the api url string
    """
    api_url = __salt__["config.get"]("mattermost.api_url") or __salt__["config.get"](
        "mattermost:api_url"
    )
    if not api_url:
        raise SaltInvocationError("No Mattermost API URL found")

    return api_url


def _get_channel():
    """
    Retrieves the Mattermost's configured channel
    :return:            String: the channel string
    """
    channel = __salt__["config.get"]("mattermost.channel") or __salt__["config.get"](
        "mattermost:channel"
    )

    return channel


def _get_username():
    """
    Retrieves the Mattermost's configured username
    :return:            String: the username string
    """
    username = __salt__["config.get"]("mattermost.username") or __salt__["config.get"](
        "mattermost:username"
    )

    return username


def post_message(message, channel=None, username=None, api_url=None, hook=None):
    """
    Send a message to a Mattermost channel.
    :param channel:     The channel name, either will work.
    :param username:    The username of the poster.
    :param message:     The message to send to the Mattermost channel.
    :param api_url:     The Mattermost api url, if not specified in the configuration.
    :param hook:        The Mattermost hook, if not specified in the configuration.
    :return:            Boolean if message was sent successfully.

    CLI Example:

    .. code-block:: bash

        salt-run mattermost.post_message message='Build is done'
    """
    if not api_url:
        api_url = _get_api_url()

    if not hook:
        hook = _get_hook()

    if not username:
        username = _get_username()

    if not channel:
        channel = _get_channel()

    if not message:
        log.error("message is a required option.")

    parameters = dict()
    if channel:
        parameters["channel"] = channel
    if username:
        parameters["username"] = username
    parameters["text"] = "```" + message + "```"  # pre-formatted, fixed-width text
    log.debug("Parameters: %s", parameters)
    data = salt.utils.json.dumps(parameters)
    result = salt.utils.mattermost.query(
        api_url=api_url, hook=hook, data=f"payload={data}"
    )

    if result:
        return True
    else:
        return result


def post_event(event, channel=None, username=None, api_url=None, hook=None):
    """
    Send an event to a Mattermost channel.
    :param channel:     The channel name, either will work.
    :param username:    The username of the poster.
    :param event:       The event to send to the Mattermost channel.
    :param api_url:     The Mattermost api url, if not specified in the configuration.
    :param hook:        The Mattermost hook, if not specified in the configuration.
    :return:            Boolean if message was sent successfully.
    """
    if not api_url:
        api_url = _get_api_url()

    if not hook:
        hook = _get_hook()

    if not username:
        username = _get_username()

    if not channel:
        channel = _get_channel()

    if not event:
        log.error("message is a required option.")

    log.debug("Event: %s", event)
    log.debug("Event data: %s", event["data"])
    message = "tag: {}\r\n".format(event["tag"])
    for key, value in event["data"].items():
        message += f"{key}: {value}\r\n"
    result = post_message(
        message, channel=channel, username=username, api_url=api_url, hook=hook
    )
    return bool(result)



import logging
import os
import tempfile

import pytest

import salt.config
import salt.utils.platform

pytestmark = [
    pytest.mark.windows_whitelisted,
]

log = logging.getLogger(__name__)


def test_minion_config_type_check(caplog):
    msg = "Config option 'ipc_write_buffer' with value"
    caplog.set_level(logging.WARNING)
    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, "w") as tmp:
            tmp.write("ipc_write_buffer: 'dynamic'\n")
        salt.config.minion_config(path)

        assert msg not in caplog.text
    finally:
        os.remove(path)


def test_cloud_config_relative_to_root_dir(tmp_path):
    root_path = tmp_path
    config_path = tmp_path / "conf"
    config_path.mkdir()
    cloud_config = config_path / "cloud"
    cloud_config.write_text("")
    master_config = config_path / "master"
    master_config.write_text(f"root_dir: {root_path}")
    opts = salt.config.cloud_config(cloud_config)
    assert opts["log_file"] == str(root_path / "var" / "log" / "salt" / "cloud")
    assert opts["cachedir"] == str(root_path / "var" / "cache" / "salt" / "cloud")


def test_master_config_relative_to_root_dir(tmp_path):
    root_path = tmp_path
    config_path = tmp_path / "conf"
    config_path.mkdir()
    master_config = config_path / "master"
    master_config.write_text(f"root_dir: {root_path}")
    opts = salt.config.master_config(master_config)
    if salt.utils.platform.is_windows():
        assert opts["pki_dir"] == str(root_path / "conf" / "pki" / "master")
    else:
        assert opts["pki_dir"] == str(root_path / "etc" / "salt" / "pki" / "master")
    assert opts["cachedir"] == str(root_path / "var" / "cache" / "salt" / "master")
    assert opts["pidfile"] == str(root_path / "var" / "run" / "salt-master.pid")
    assert opts["sock_dir"] == str(root_path / "var" / "run" / "salt" / "master")
    assert opts["extension_modules"] == str(
        root_path / "var" / "cache" / "salt" / "master" / "extmods"
    )
    assert opts["token_dir"] == str(
        root_path / "var" / "cache" / "salt" / "master" / "tokens"
    )
    assert opts["syndic_dir"] == str(
        root_path / "var" / "cache" / "salt" / "master" / "syndics"
    )
    assert opts["sqlite_queue_dir"] == str(
        root_path / "var" / "cache" / "salt" / "master" / "queues"
    )
    assert opts["log_file"] == str(root_path / "var" / "log" / "salt" / "master")
    assert opts["key_logfile"] == str(root_path / "var" / "log" / "salt" / "key")
    assert opts["ssh_log_file"] == str(root_path / "var" / "log" / "salt" / "ssh")

    # These are not tested because we didn't define them in the master config.
    # assert opts["autosign_file"] == str(root_path / "var" / "run" / "salt"/ "master")
    # assert opts["autoreject_file"] == str(root_path / "var" / "run" / "salt"/ "master")
    # assert opts["autosign_grains_dir"] == str(root_path / "var" / "run" / "salt"/ "master")


def test_minion_config_relative_to_root_dir(tmp_path):
    root_path = tmp_path
    config_path = tmp_path / "conf"
    config_path.mkdir()
    minion_config = config_path / "minion"
    minion_config.write_text(f"root_dir: {root_path}")
    opts = salt.config.minion_config(minion_config)
    if salt.utils.platform.is_windows():
        assert opts["pki_dir"] == str(root_path / "conf" / "pki" / "minion")
    else:
        assert opts["pki_dir"] == str(root_path / "etc" / "salt" / "pki" / "minion")
    assert opts["cachedir"] == str(root_path / "var" / "cache" / "salt" / "minion")
    assert opts["pidfile"] == str(root_path / "var" / "run" / "salt-minion.pid")
    assert opts["sock_dir"] == str(root_path / "var" / "run" / "salt" / "minion")
    assert opts["extension_modules"] == str(
        root_path / "var" / "cache" / "salt" / "minion" / "extmods"
    )
    assert opts["log_file"] == str(root_path / "var" / "log" / "salt" / "minion")


def test_api_config_relative_to_root_dir(tmp_path):
    root_path = tmp_path
    config_path = tmp_path / "conf"
    config_path.mkdir()
    master_config = config_path / "master"
    master_config.write_text(f"root_dir: {root_path}")
    opts = salt.config.api_config(master_config)
    assert opts["pidfile"] == str(root_path / "var" / "run" / "salt-api.pid")
    assert opts["log_file"] == str(root_path / "var" / "log" / "salt" / "api")
    assert opts["api_pidfile"] == str(root_path / "var" / "run" / "salt-api.pid")
    assert opts["api_logfile"] == str(root_path / "var" / "log" / "salt" / "api")


def test_spm_config_relative_to_root_dir(tmp_path):
    root_path = tmp_path
    config_path = tmp_path / "conf"
    config_path.mkdir()
    spm_config = config_path / "spm"
    spm_config.write_text(f"root_dir: {root_path}")
    opts = salt.config.spm_config(spm_config)

    assert opts["formula_path"] == str(root_path / "srv" / "spm" / "salt")
    assert opts["pillar_path"] == str(root_path / "srv" / "spm" / "pillar")
    assert opts["reactor_path"] == str(root_path / "srv" / "spm" / "reactor")
    assert opts["spm_cache_dir"] == str(root_path / "var" / "cache" / "salt" / "spm")
    assert opts["spm_build_dir"] == str(root_path / "srv" / "spm_build")
    assert opts["spm_logfile"] == str(root_path / "var" / "log" / "salt" / "spm")


def test_syndic_config_relative_to_root_dir(tmp_path):
    root_path = tmp_path
    config_path = tmp_path / "conf"
    config_path.mkdir()
    master_config = config_path / "master"
    master_config.write_text(f"root_dir: {root_path}")
    minion_config = config_path / "master"
    minion_config.write_text(f"root_dir: {root_path}")
    opts = salt.config.syndic_config(master_config, minion_config)
    if salt.utils.platform.is_windows():
        assert opts["pki_dir"] == str(root_path / "conf" / "pki" / "minion")
    else:
        assert opts["pki_dir"] == str(root_path / "etc" / "salt" / "pki" / "minion")
    assert opts["cachedir"] == str(root_path / "var" / "cache" / "salt" / "master")
    assert opts["pidfile"] == str(root_path / "var" / "run" / "salt-syndic.pid")
    assert opts["sock_dir"] == str(root_path / "var" / "run" / "salt" / "minion")
    assert opts["extension_modules"] == str(
        root_path / "var" / "cache" / "salt" / "minion" / "extmods"
    )
    assert opts["token_dir"] == str(
        root_path / "var" / "cache" / "salt" / "master" / "tokens"
    )
    assert opts["log_file"] == str(root_path / "var" / "log" / "salt" / "syndic")
    assert opts["key_logfile"] == str(root_path / "var" / "log" / "salt" / "key")
    assert opts["syndic_log_file"] == str(root_path / "var" / "log" / "salt" / "syndic")



"""
Connection module for Telemetry

.. versionadded:: 2016.3.0

https://github.com/mongolab/mongolab-telemetry-api-docs/blob/master/alerts.md

:configuration: This module accepts explicit telemetry credentials or
    can also read api key credentials from a pillar. More Information available
    here__.

.. __: https://github.com/mongolab/mongolab-telemetry-api-docs/blob/master/alerts.md

In the minion's config file:

.. code-block:: yaml

   telemetry.telemetry_api_keys:
     - abc123  # Key 1
     - efg321  # Backup Key 1
   telemetry_api_base_url: https://telemetry-api.mongolab.com/v0

:depends: requests

"""

import logging

import salt.utils.json
import salt.utils.stringutils

try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False

log = logging.getLogger(__name__)

__virtualname__ = "telemetry"


def __virtual__():
    # Only load if imports exist.
    if not HAS_REQUESTS:
        return False
    return __virtualname__


def _get_telemetry_base(profile):
    config = __salt__["config.option"](profile)
    return config.get("telemetry_api_base_url")


def _auth(api_key=None, profile="telemetry"):
    # return telemetry api key in the header
    if api_key is None and profile is None:
        raise Exception("Missing api_key and profile")
    if profile:
        if isinstance(profile, str):
            _profile = __salt__["config.option"](profile)
        elif isinstance(profile, dict):
            _profile = profile

        if _profile:
            api_key = _profile.get("telemetry_api_keys")[0]
        else:
            raise Exception("Missing api_key")

    return {"Telemetry-API-Key": api_key, "content-type": "application/json"}


def _update_cache(deployment_id, metric_name, alert):
    key = f"telemetry.{deployment_id}.alerts"

    if key in __context__:
        alerts = __context__[key]
        alerts[metric_name] = alert
        __context__[key] = alerts

    return __context__.get(key, [])


def _retrieve_channel_id(email, profile="telemetry"):
    """
    Given an email address, checks the local
    cache if corresponding email address: channel_id
    mapping exists

    email
        Email escalation policy
    profile
        A dict of telemetry config information.
    """
    key = "telemetry.channels"
    auth = _auth(profile=profile)

    if key not in __context__:
        get_url = (
            _get_telemetry_base(profile)
            + "/notification-channels?_type=EmailNotificationChannel"
        )
        response = requests.get(get_url, headers=auth, timeout=120)

        if response.status_code == 200:
            cache_result = {}
            for index, alert in enumerate(response.json()):
                cache_result[alert.get("email")] = alert.get("_id", "false")

            __context__[key] = cache_result

    return __context__[key].get(email, False)


def get_alert_config(
    deployment_id, metric_name=None, api_key=None, profile="telemetry"
):
    """
    Get all alert definitions associated with a given deployment or if metric_name
    is specified, obtain the specific alert config

    Returns dictionary or list of dictionaries.

    CLI Example:

    .. code-block:: bash

        salt myminion telemetry.get_alert_config rs-ds033197 currentConnections profile=telemetry
        salt myminion telemetry.get_alert_config rs-ds033197 profile=telemetry
    """

    auth = _auth(profile=profile)
    alert = False

    key = f"telemetry.{deployment_id}.alerts"

    if key not in __context__:
        try:
            get_url = _get_telemetry_base(profile) + "/alerts?deployment={}".format(
                deployment_id
            )
            response = requests.get(get_url, headers=auth, timeout=120)
        except requests.exceptions.RequestException as e:
            log.error(str(e))
            return False

        http_result = {}
        if response.status_code == 200:
            for alert in response.json():
                http_result[alert.get("condition", {}).get("metric")] = alert
                __context__[key] = http_result

    if not __context__.get(key):
        return []

    alerts = __context__[key].values()

    if metric_name:
        return __context__[key].get(metric_name)

    return [alert["_id"] for alert in alerts if "_id" in alert]


def get_notification_channel_id(notify_channel, profile="telemetry"):
    """
    Given an email address, creates a notification-channels
    if one is not found and also returns the corresponding
    notification channel id.

    notify_channel
        Email escalation policy
    profile
        A dict of telemetry config information.

    CLI Example:

    .. code-block:: bash

        salt myminion telemetry.get_notification_channel_id userx@company.com profile=telemetry
    """

    # This helper is used to procure the channel ids
    # used to notify when the alarm threshold is violated
    auth = _auth(profile=profile)

    notification_channel_id = _retrieve_channel_id(notify_channel)

    if not notification_channel_id:
        log.info("%s channel does not exist, creating.", notify_channel)

        # create the notification channel and cache the id
        post_url = _get_telemetry_base(profile) + "/notification-channels"
        data = {
            "_type": "EmailNotificationChannel",
            "name": notify_channel[: notify_channel.find("@")] + "EscalationPolicy",
            "email": notify_channel,
        }
        response = requests.post(
            post_url, data=salt.utils.json.dumps(data), headers=auth, timeout=120
        )
        if response.status_code == 200:
            log.info(
                "Successfully created EscalationPolicy %s with"
                " EmailNotificationChannel %s",
                data.get("name"),
                notify_channel,
            )
            notification_channel_id = response.json().get("_id")
            __context__["telemetry.channels"][notify_channel] = notification_channel_id
        else:
            raise Exception(f"Failed to created notification channel {notify_channel}")

    return notification_channel_id


def get_alarms(deployment_id, profile="telemetry"):
    """
    get all the alarms set up against the current deployment

    Returns dictionary of alarm information

    CLI Example:

    .. code-block:: bash

        salt myminion telemetry.get_alarms rs-ds033197 profile=telemetry

    """
    auth = _auth(profile=profile)

    try:
        response = requests.get(
            _get_telemetry_base(profile) + f"/alerts?deployment={deployment_id}",
            headers=auth,
            timeout=120,
        )
    except requests.exceptions.RequestException as e:
        log.error(str(e))
        return False

    if response.status_code == 200:
        alarms = response.json()

        if alarms:
            return alarms

        return f"No alarms defined for deployment: {deployment_id}"
    else:
        # Non 200 response, sent back the error response'
        return {
            "err_code": response.status_code,
            "err_msg": salt.utils.json.loads(response.text).get("err", ""),
        }


def create_alarm(deployment_id, metric_name, data, api_key=None, profile="telemetry"):
    """
    create an telemetry alarms.

    data is a dict of alert configuration data.

    Returns (bool success, str message) tuple.

    CLI Example:

    .. code-block:: bash

        salt myminion telemetry.create_alarm rs-ds033197 {} profile=telemetry

    """

    auth = _auth(api_key, profile)
    request_uri = _get_telemetry_base(profile) + "/alerts"

    key = f"telemetry.{deployment_id}.alerts"

    # set the notification channels if not already set
    post_body = {
        "deployment": deployment_id,
        "filter": data.get("filter"),
        "notificationChannel": get_notification_channel_id(
            data.get("escalate_to")
        ).split(),
        "condition": {
            "metric": metric_name,
            "max": data.get("max"),
            "min": data.get("min"),
        },
    }

    try:
        response = requests.post(
            request_uri,
            data=salt.utils.json.dumps(post_body),
            headers=auth,
            timeout=120,
        )
    except requests.exceptions.RequestException as e:
        # TODO: May be we should retry?
        log.error(str(e))

    if response.status_code >= 200 and response.status_code < 300:
        # update cache
        log.info(
            "Created alarm on metric: %s in deployment: %s", metric_name, deployment_id
        )
        log.debug(
            "Updating cache for metric %s in deployment %s: %s",
            metric_name,
            deployment_id,
            response.json(),
        )
        _update_cache(deployment_id, metric_name, response.json())
    else:
        log.error(
            "Failed to create alarm on metric: %s in deployment %s: payload: %s",
            metric_name,
            deployment_id,
            salt.utils.json.dumps(post_body),
        )

    return response.status_code >= 200 and response.status_code < 300, response.json()


def update_alarm(deployment_id, metric_name, data, api_key=None, profile="telemetry"):
    """
    update an telemetry alarms. data is a dict of alert configuration data.

    Returns (bool success, str message) tuple.

    CLI Example:

    .. code-block:: bash

        salt myminion telemetry.update_alarm rs-ds033197 {} profile=telemetry

    """
    auth = _auth(api_key, profile)
    alert = get_alert_config(deployment_id, metric_name, api_key, profile)

    if not alert:
        return (
            False,
            "No entity found matching deployment {} and alarms {}".format(
                deployment_id, metric_name
            ),
        )

    request_uri = _get_telemetry_base(profile) + "/alerts/" + alert["_id"]

    # set the notification channels if not already set
    post_body = {
        "deployment": deployment_id,
        "filter": data.get("filter"),
        "notificationChannel": get_notification_channel_id(
            data.get("escalate_to")
        ).split(),
        "condition": {
            "metric": metric_name,
            "max": data.get("max"),
            "min": data.get("min"),
        },
    }

    try:
        response = requests.put(
            request_uri,
            data=salt.utils.json.dumps(post_body),
            headers=auth,
            timeout=120,
        )
    except requests.exceptions.RequestException as e:
        log.error("Update failed: %s", e)
        return False, str(e)

    if response.status_code >= 200 and response.status_code < 300:
        # Also update cache
        log.debug(
            "Updating cache for metric %s in deployment %s: %s",
            metric_name,
            deployment_id,
            response.json(),
        )
        _update_cache(deployment_id, metric_name, response.json())
        log.info(
            "Updated alarm on metric: %s in deployment: %s", metric_name, deployment_id
        )
        return True, response.json()

    err_msg = (
        "Failed to create alarm on metric: {} in deployment: {} payload: {}".format(
            salt.utils.stringutils.to_unicode(metric_name),
            salt.utils.stringutils.to_unicode(deployment_id),
            salt.utils.json.dumps(post_body),
        )
    )
    log.error(err_msg)
    return False, err_msg


def delete_alarms(
    deployment_id, alert_id=None, metric_name=None, api_key=None, profile="telemetry"
):
    """delete an alert specified by alert_id or if not specified blows away all the alerts
         in the current deployment.

    Returns (bool success, str message) tuple.

    CLI Example:

    .. code-block:: bash

        salt myminion telemetry.delete_alarms rs-ds033197 profile=telemetry

    """
    auth = _auth(profile=profile)

    if alert_id is None:
        # Delete all the alarms associated with this deployment
        alert_ids = get_alert_config(deployment_id, api_key=api_key, profile=profile)
    else:
        alert_ids = [alert_id]

    if not alert_ids:
        return (
            False,
            f"failed to find alert associated with deployment: {deployment_id}",
        )

    failed_to_delete = []
    for id in alert_ids:
        delete_url = _get_telemetry_base(profile) + f"/alerts/{id}"

        try:
            response = requests.delete(delete_url, headers=auth, timeout=120)
            if metric_name:
                log.debug(
                    "updating cache and delete %s key from %s",
                    metric_name,
                    deployment_id,
                )
                _update_cache(deployment_id, metric_name, None)

        except requests.exceptions.RequestException as e:
            log.error("Delete failed: %s", e)

        if response.status_code != 200:
            failed_to_delete.append(id)

    if failed_to_delete:
        return (
            False,
            "Failed to delete {} alarms in deployment: {}".format(
                ", ".join(failed_to_delete), deployment_id
            ),
        )

    return (
        True,
        "Successfully deleted {} alerts in deployment: {}".format(
            ", ".join(alert_ids), deployment_id
        ),
    )
